import numpy as np
import matplotlib.pyplot as plt
import scipy as sp


def KL(a, b):
    # a = np.asarray(a, dtype=np.float)
    # b = np.asarray(b, dtype=np.float)

    return np.sum(np.where(a != 0, a * np.log(a / b), 0))

def read_spikes(fn):

    X = np.loadtxt(fn, delimiter=',',skiprows=1)
    T = np.copy(X[:,1])
    Y = np.copy(X[:,0])
    Y = Y.astype(int)
	return T, Y


def calc_ISIs(T, Y):
	M = int(np.max(Y) + 1)
	ISIs = []
	for i in range(M):
		Tsub = T[Y == i]
		if len(Tsub) > 0:
			Tsub = np.sort(Tsub)
			DT = Tsub[1:] - Tsub[:-1]
			ISIs.append(DT)
		else:
			ISIs.append([])
	return ISIs


def calc_CV(T, Y):
	P = calc_ISIs(T, Y)
	mu = np.zeros(len(P))
	var = np.zeros(len(P))
	CV = np.zeros(len(P))
	for i in range(len(P)):
		mu[i] = np.mean(P[i]) if len(P[i]) > 0 else -1
		var[i] = np.var(P[i]) if len(P[i]) > 0 else -1
		CV[i] = var[i] / mu[i] if mu[i] > 0 else -1
	return mu, var, CV


methods = ['true', 'MLE', 'MAP', 'GML']

mu=[]
var=[]
CV=[]

fn = 'control_data_0_spikes_newid.csv'
T, Y = read_spikes(fn)
mu_true, var_true, CV_true = calc_CV(T, Y)
mu.append(mu_true)
var.append(var_true)
CV.append(CV_true)

fn = 'control_data_0_monte_spikes_MLE.csv'
T, Y = read_spikes(fn)
mu_MLE, var_MLE, CV_MLE = calc_CV(T, Y)
mu.append(mu_MLE)
var.append(var_MLE)
CV.append(CV_MLE)

fn = 'control_data_0_monte_spikes_MAP.csv'
T, Y = read_spikes(fn)
mu_MAP, var_MAP, CV_MAP = calc_CV(T, Y)
mu.append(mu_MAP)
var.append(var_MAP)
CV.append(CV_MAP)

fn = 'control_data_0_monte_spikes_ML-II.csv'
T, Y = read_spikes(fn)
mu_GML, var_GML, CV_GML = calc_CV(T, Y)
mu.append(mu_GML)
var.append(var_GML)
CV.append(CV_GML)





mu_dist = [np.histogram(mu[i], bins=20, range=(np.min(mu), np.max(mu)), density=True) for i in range(len(mu))]
CV_dist = [np.histogram(CV[i], bins=20, range=(np.min(CV), np.max(CV)), density=True) for i in range(len(CV))]





noise = np.random.rand()*1e-8
KL_mu = [ KL(mu_dist[i][0]+noise, mu_dist[0][0]+noise) for i in range(len(mu_dist))]





noise = np.random.rand()*1e-8
KL_CV = [ KL(CV_dist[i][0]+noise, CV_dist[0][0]+noise) for i in range(len(CV_dist))]

np.savetxt('ISIs_mu_KL.txt', KL_mu, delimiter=',')
np.savetxt('ISIs_CV_KL.txt', KL_CV, delimiter=',')






plt.close()
for i in range(len(mu_dist)):
	plt.plot((mu_dist[i][1][:-1]+mu_dist[i][1][1:])/2, mu_dist[i][0], label = methods[i])
plt.legend(loc='best')
# plt.show()
plt.savefig('ISIs_mu.pdf')



plt.close()
for i in range(len(CV_dist)):
	plt.plot((CV_dist[i][1][:-1]+CV_dist[i][1][1:])/2, CV_dist[i][0], label = methods[i])
plt.legend(loc='best')
# plt.show()
plt.savefig('ISIs_CV.pdf')




