import math
import numpy as np
import matplotlib.pyplot as plt
import scipy
import scipy.integrate as integrate
from scipy import special
from scipy.stats import norm

# import sklearn
import multiprocessing
from multiprocessing import Pool
import itertools
from matplotlib.colors import LogNorm
import matplotlib
from matplotlib.patches import Patch
from matplotlib.lines import Line2D

import os
from joblib import dump, load


# import ndd

import sys
import time

from numba import jit, njit
import numba

rescale_pos_ROI = True

import numba as nb
import numpy as np


@numba.vectorize(['float64(float64)'], target='cpu', nopython=True, fastmath=True)
def exp_nb(x):
	return np.exp(x)

@numba.vectorize(['float64(float64)'], target='cpu', nopython=True, fastmath=True)
def log_nb(x):
	return np.log(x)

@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def exp_custom(x):
	return exp_nb(x)

@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def log_custom(x):
	return log_nb(x)

@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def tanh_custom(x):
	return (1 - 2/(exp_nb(2*x) + 1))


def flatten(t):
	return [item for sublist in t for item in sublist]

def filter_data(dataset , col_list, var_list):
	loc = [ set(np.where(dataset[:,col_list[i]]==var_list[i])[0]) for i in range(len(col_list))]
	loc = np.array(list(set.intersection(*loc)))
	loc = np.unique(loc)
	# return dataset[loc,:]
	return loc

def replace_time_intervals(event_id, event_time, gap_size, new_gap_size):
	new_time = [event_time[0]]
	for i in range(1, len(event_time)):
		if event_time[i] - event_time[i-1] > gap_size:
			new_time.append(new_time[-1] + new_gap_size)
		else:
			new_time.append(event_time[i])
	return new_time


def replace_consecutive_zero_rows(X):
	"""
	Erases consecutive rows that are all zeros in every matrix s in X
	and resizes the matrix accordingly.

	Args:
	- X: list of numpy arrays, each s in X is a T by N matrix

	Returns:
	- X: list of numpy arrays with modified matrices
	"""
	# Assume X is a list of matrices
	min_len = 0
	for i, s in enumerate(X):
		# print(len(s))
		# np.savetxt('before{}.txt'.format(i), s, delimiter='\t')
		# fig, ax = plt.subplots(figsize=(20, len(s)))
		# ax.matshow(s)
		# plt.savefig('before{}.pdf'.format(i))
		# Find consecutive rows of zeros
		zero_rows = np.all(s == -1, axis=1)
		zero_row_starts = np.where(np.diff(zero_rows.astype(int)) == 1)[0] + 1
		zero_row_ends = np.where(np.diff(zero_rows.astype(int)) == -1)[0] + 1
		if zero_rows[0]:
			zero_row_starts = np.concatenate([[0], zero_row_starts])
		if zero_rows[-1]:
			zero_row_ends = np.concatenate([zero_row_ends, [s.shape[0]]])
		zero_row_lengths = zero_row_ends - zero_row_starts

		# Erase consecutive rows of zeros
		for start, length in zip(zero_row_starts, zero_row_lengths):
			if length > 1:
				# print(start, start+length)
				s[start:start + length-1, :] = 0
		s = s[~np.all(s == 0, axis=1)]

		# plt.close()
		# fig, ax = plt.subplots(figsize=(20, len(s)))
		# ax.matshow(s)
		# plt.savefig('after{}.pdf'.format(i))
		# plt.close()
		# np.savetxt('after{}.txt'.format(i), s, delimiter='\t')
		# print(len(s))
		# print('\n')
		X[i] = s
		if min_len == 0:
			min_len = len(s)
		else:
			if len(s) < min_len:
				min_len = len(s)
				for j in range(i):
					X[j] = X[j][:min_len, :]
			else:
				X[i] = X[i][:min_len, :]


	return X


def generate_random_regular_graph(NodeNo, K):
	# NodeNo = 100
	# K = 3
	A_ij = np.zeros( (NodeNo, NodeNo) )
	neighbor_i = np.ones( (NodeNo, K) )*-1
	c_i = np.zeros( NodeNo ).astype(int)
	if_valid = 1
	for i in range(NodeNo):
		rand_list = np.array(range(NodeNo))[c_i<K]
		np.random.shuffle(rand_list)
		rand_list
		counter = 0
		while(c_i[i]<K and if_valid):
			if counter < len(rand_list):
				if(A_ij[i][ rand_list[counter] ] == 0 and rand_list[counter] != i):
					A_ij[i][ rand_list[counter] ] = 1
					A_ij[ rand_list[counter] ][i] = 1
					neighbor_i[i][c_i[i]] = rand_list[counter]
					neighbor_i[rand_list[counter]][c_i[rand_list[counter]]] = i
					c_i[i] += 1
					c_i[rand_list[counter]] += 1
				counter += 1
			else:
				if_valid = 0
		# print(i)
	if_connected = 0
	if ((np.linalg.matrix_power(A_ij, NodeNo) == 0).sum(axis=0)==0).sum() > 0:
		if_connected = 1
	if(if_valid and if_connected):
		return A_ij,neighbor_i, c_i
	else:
		# print(if_valid , if_connected)
		return generate_random_regular_graph(NodeNo, K)

def generate_directed_random_regular_graph_neural(NodeNo, para, allow_self_loop):
	# NodeNo = 100
	# K = 3
	K = int(para)
	A_ij = np.zeros( (NodeNo, NodeNo) )
	neighbor_i = np.ones( (NodeNo, K) )*-1
	c_i = np.zeros( NodeNo ).astype(int)
	if_valid = 1
	for i in range(NodeNo):
		rand_list = np.array(range(NodeNo))
		if not allow_self_loop:
			z = (rand_list == i)
			rand_list = rand_list[~z]
		np.random.shuffle(rand_list)
		A_ij[i][rand_list[:K]] = 1
		neighbor_i[i] = rand_list[:K]
		c_i[i] = K
	return A_ij,neighbor_i, c_i

def generate_RRG_with_loc_neural(NodeNo, para, allow_self_loop):
	loc = np.zeros((NodeNo,2))
	posx = np.random.rand(NodeNo)
	posy = np.random.rand(NodeNo)
	loc[:,0] = posx
	loc[:,1] = posy
	dis_ij = np.array([[ ((posx[i]-posx[j])**2 + (posy[i]-posy[j])**2)**(0.5)  for j in range(NodeNo) ] for i in range(NodeNo)])
	p_ij = exp_custom(-para*NodeNo*dis_ij)
	if not allow_self_loop:
		p_ij[ np.diag_indices(len(p_ij)) ] = 0
	tmp_P = np.random.rand( NodeNo, NodeNo)
	A_ij = np.zeros( (NodeNo, NodeNo) )
	A_ij[tmp_P < p_ij] = 1
	if not allow_self_loop:
		for i in range(len(A_ij)):
			A_ij[i,i] = 0
	neighbor_i = np.ones( (NodeNo, NodeNo) )*-1
	c_i = np.zeros( NodeNo ).astype(int)
	return A_ij, neighbor_i, c_i, dis_ij, loc


def generate_data_guassian_mixture(ratios, means, sds, N):

	no_of_mixtures = len(ratios)
	sizei = (np.array(ratios)*np.sqrt(N)).astype(int)
	caltmp = np.array( [] )

	i=0
	rand_tmp = -np.exp(np.random.normal( (means[i]) , sds[i], size=int( sizei[i]*np.sqrt(N) ) ))
	caltmp = np.append( caltmp, rand_tmp )
	i=1
	rand_tmp = np.exp(np.random.normal( (means[i]) , sds[i], size=int( sizei[i]*np.sqrt(N) ) ))
	caltmp = np.append( caltmp, rand_tmp )

	n = int(np.sqrt(N))
	caltmp = np.copy(caltmp.reshape((n,n)))

	for i in range(n):
		caltmp[i,i] = -np.exp(np.random.normal( means[0]/3 , sds[0] ))

	return caltmp


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def get_m_C_D_from_S(S):
	number_of_sample = len(S)
	time_step = len(S[0])
	N = len(S[0,0])

	m_i = np.zeros((number_of_sample, N))
	C = np.zeros((number_of_sample, N,N))
	D = np.zeros((number_of_sample, N,N))

	for r in range(number_of_sample):
		for t in range(time_step-1):
			m_i[r] += S[r, t+1, :]
			tmpC = np.array( [[(S[r, t+1, i] * S[r, t+1, j]) for j in range(N)] for i in range(N)] )
			C[r] += tmpC
			tmpD = np.array( [[(S[r, t+1, i] * S[r, t+0, j]) for j in range(N)] for i in range(N)] )
			D[r] += tmpD

		for i in range(N):
			m_i[r,i] = m_i[r,i]/(time_step-1)
		for i in range(N):
			for j in range(N):
				C[r,i,j] = C[r,i,j]/((time_step-1)) - m_i[r,i]*m_i[r,j]
				D[r,i,j] = D[r,i,j]/((time_step-1)) - m_i[r,i]*m_i[r,j]

	return m_i, C, D




def log_likelihood(S, J, H):
	no_R = len(S)
	no_T = len(S[0])-1
	no_N = len(S[0,0,:])

	M_tmp = np.array( [S[:,1:,i].mean() for i in range(no_N)] )
	D_tmp = np.array( [[(S[:,1:,i] * S[:,:-1,j]).mean() for j in range(no_N)] for i in range(no_N)] )
	T3 = np.array( [log_custom( 2*np.cosh( (np.dot( J, S[r].T ).T + H).T ) ).mean(axis=1) for r in range(no_R)] ).mean(axis=0)

	log_L = no_T*no_R*(H*M_tmp + (J*D_tmp).sum(axis=1) - T3).sum()

	return log_L

def log_posteriori(S, J, H, Prior_J, Prior_H):

	log_L = log_likelihood(S, J, H)
	# print(log_L)

	Prior_J_omega = Prior_J[0]
	Prior_J_mu = Prior_J[1]
	Prior_J_sigma = Prior_J[2]

	Prior_H_mu = Prior_H[0]
	Prior_H_sigma = Prior_H[1]

	delta_J = np.zeros((len(J), len(J)))
	delta_J[J==0] = 1
	p_P= (delta_J*Prior_J_omega + (1-Prior_J_omega)*np.exp((-1*(J-Prior_J_mu)**2)/(2*(Prior_J_sigma**2)))/(Prior_J_sigma*np.sqrt(2*np.pi)))
	# p_P[p_P < 1e-20] = 0
	# p_P[p_P == 0] = 1e-20
	# if( (p_P == 0).sum()>0 ):
	# 	return -np.inf
	log_p_P = np.copy(p_P)
	log_p_P[p_P==0] = -1e-20
	log_p_P[p_P>0] = log_custom(p_P[p_P>0])

	log_p_H = -1*(H - Prior_H_mu)**2/(2*(Prior_H_sigma**2)) - log_custom(Prior_H_sigma*np.sqrt(2*np.pi))

	caltmp = log_L + log_p_P.sum() + log_p_H.sum()
	# print(caltmp)

	return caltmp

@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def gaussian_pdf(x, mu, sd):
	var = sd**2
	root_2pi_var = sd * ((2*np.pi)**0.5)
	return exp_custom(  (x - mu)**2/(-2*(var))  )/root_2pi_var

@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def P_z_j_one_j(input_J_grad, all_gaussians, prior_theta, prior_e_minus_a_l_ij, priors_J, which_j, small_epsilon):

	gamma = priors_J[0]
	mus = np.array([0, priors_J[1], priors_J[2] ])
	sds = np.array([small_epsilon, priors_J[3]+small_epsilon, priors_J[4]+small_epsilon ])


	root_2pi_var = sds * ((2*np.pi)**0.5)

	J_lj_s = np.copy(input_J_grad[:,which_j])
	list_tmp = (np.arange(len(J_lj_s)))
	list_tmp = np.copy(list_tmp!=which_j)
	J_lj_s = np.copy(J_lj_s[list_tmp])
	gaussians_J_lj = np.zeros((2,2,len(J_lj_s)))


	if all_gaussians == None:
		gaussians_J_lj[0,0,:] = exp_custom(  (J_lj_s - mus[0])**2/(-2*(sds[0]**2))  )/root_2pi_var[0]
		gaussians_J_lj[0,1,:] = np.copy(gaussians_J_lj[0,0,:])
		for i in range(len(mus)-1):
			for l in range(len(J_lj_s)):
				if J_lj_s[l]*(i*2-1) <0:
					gaussians_J_lj[1, i , l ] =  0
				else:
					gaussians_J_lj[1, i , l ] = exp_custom(  (log_custom( np.abs(J_lj_s[l]) ) - (mus[1+i]) )**2/(-2*(sds[1+i]**2))  )/root_2pi_var[1+i]

	else:
		gaussians_J_lj[0,0,:] = all_gaussians[0,0, :, which_j]
		gaussians_J_lj[0,1,:] = all_gaussians[0,1, :, which_j]
		gaussians_J_lj[1,0,:] = all_gaussians[1,0, :, which_j]
		gaussians_J_lj[1,1,:] = all_gaussians[1,1, :, which_j]

	P_z_j = np.zeros(2)

	z = 0; caltmp0 = (1-prior_theta[list_tmp,which_j]*prior_e_minus_a_l_ij[list_tmp,which_j])*gaussians_J_lj[0,z,:] + prior_theta[list_tmp,which_j]*prior_e_minus_a_l_ij[list_tmp,which_j]*gaussians_J_lj[1,z,:]
	z = 1; caltmp1 = (1-prior_theta[list_tmp,which_j]*prior_e_minus_a_l_ij[list_tmp,which_j])*gaussians_J_lj[0,z,:] + prior_theta[list_tmp,which_j]*prior_e_minus_a_l_ij[list_tmp,which_j]*gaussians_J_lj[1,z,:]

	if_zero_0_tmp = np.sum(caltmp0==0)
	if_zero_1_tmp = np.sum(caltmp1==0)
	if not (if_zero_0_tmp + if_zero_1_tmp == 0):
		if (not if_zero_0_tmp ==0 ) and (not if_zero_1_tmp == 0):
			return P_z_j
		else:
			if (not if_zero_0_tmp ==0 ):
				P_z_j[0] = 0.01
				P_z_j[1] = 1-P_z_j[0]
				return P_z_j
			if (not if_zero_1_tmp ==0 ):
				P_z_j[0] = 0.99
				P_z_j[1] = 1-P_z_j[0]
				return P_z_j

	ln_r = (log_custom(caltmp0/caltmp1)).sum()
	caltmp = ln_r + log_custom( gamma / (1-gamma) )
	if caltmp > 5e1:
		caltmp = 5e1
	else:
		if caltmp < -5e1:
			caltmp = -5e1
	r = exp_custom( caltmp )

	P_z_j[0] = r/(1+r)
	P_z_j[1] = 1 - P_z_j[0]

	return P_z_j

@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def P_phi_ij_given_z(input_J_grad, prior_theta, prior_e_n_a_l_ij, priors_J, which_ij_safe, small_epsilon):

	gamma = priors_J[0]
	mus = np.array([0, priors_J[1], priors_J[2] ])
	sds = np.array([small_epsilon, priors_J[3]+small_epsilon, priors_J[4]+small_epsilon ])

	P_phi_given_z = np.zeros((2,2))

	if(  which_ij_safe[0] ==  which_ij_safe[1]):
		P_phi_given_z[1,:] = 1
		return P_phi_given_z

	J_ij_tmp = input_J_grad[ which_ij_safe[0], which_ij_safe[1] ]
	ln_J_ij_tmp = log_custom( np.abs(J_ij_tmp) )

	ln_P_phi_given_z = np.zeros((2,2))
	for z in [0,1]:
		bool_tmp = False
		for phi in [0,1]:
			mu_tmp = (1-phi)*mus[0] + phi*( (1-z)*mus[1] + z*mus[2] )
			sd_tmp = (1-phi)*sds[0] + phi*( (1-z)*sds[1] + z*sds[2] )
			p_tmp = (1-phi)*(1-prior_theta[ which_ij_safe[0], which_ij_safe[1] ]*prior_e_n_a_l_ij[ which_ij_safe[0], which_ij_safe[1] ]) + phi*prior_theta[ which_ij_safe[0], which_ij_safe[1] ]*prior_e_n_a_l_ij[ which_ij_safe[0], which_ij_safe[1] ]
			if phi == 0:
				ln_P_phi_given_z[phi,z] = ( -(J_ij_tmp - mu_tmp)**2/(2* (sd_tmp**2) ) ) - 0.5*log_custom(2*np.pi*sd_tmp*sd_tmp) + log_custom(p_tmp)
			else:
				if J_ij_tmp*(z*2-1) < 0:
					ln_P_phi_given_z[phi,z] = 0
					bool_tmp = True
				else:
					ln_P_phi_given_z[phi,z] = ( -( ln_J_ij_tmp - (mu_tmp))**2/(2* (sd_tmp**2) ) ) - 0.5*log_custom(2*np.pi*sd_tmp*sd_tmp) + log_custom(p_tmp)

		if not bool_tmp:
			ln_r = ln_P_phi_given_z[0,z]-ln_P_phi_given_z[1,z]
			if ln_r > 5e1:
				ln_r = 5e1
			else:
				if ln_r < -5e1:
					ln_r = -5e1
			r = exp_custom( ln_r )
			P_phi_given_z[0,z] = r/(1+r)
			P_phi_given_z[1,z] = 1-P_phi_given_z[0,z]
		else:
			P_phi_given_z[0,z] = 1-0.01
			P_phi_given_z[1,z] = 0+0.01

	return P_phi_given_z

#
#
# @njit(nogil=True, cache=True, fastmath=True, looplift=True)
# def P_phi_ij(input_J_grad, prior_theta, prior_e_n_a_l_ij, priors_J, which_ij_safe, small_epsilon):
# 	gamma = priors_J[0]
# 	mus = np.array([0, priors_J[1], priors_J[2] ])
# 	sds = np.array([small_epsilon, priors_J[3]+small_epsilon, priors_J[4]+small_epsilon ])
#
# 	P_phi_given_z = np.zeros((2,2))
#
# 	if(  which_ij_safe[0] ==  which_ij_safe[1]):
# 		P_phi_given_z[1,:] = 1
# 		return P_phi_given_z
#
# 	J_ij_tmp = input_J_grad[ which_ij_safe[0], which_ij_safe[1] ]
# 	ln_J_ij_tmp = log_custom( np.abs(J_ij_tmp) )
#
# 	ln_P_phi_given_z = np.zeros((2,2))
# 	for z in [0,1]:
# 		bool_tmp = False
# 		for phi in [0,1]:
# 			mu_tmp = (1-phi)*mus[0] + phi*( (1-z)*mus[1] + z*mus[2] )
# 			sd_tmp = (1-phi)*sds[0] + phi*( (1-z)*sds[1] + z*sds[2] )
# 			p_tmp = (1-phi)*(1-prior_theta[ which_ij_safe[0], which_ij_safe[1] ]*prior_e_n_a_l_ij[ which_ij_safe[0], which_ij_safe[1] ]) + phi*prior_theta[ which_ij_safe[0], which_ij_safe[1] ]*prior_e_n_a_l_ij[ which_ij_safe[0], which_ij_safe[1] ]
# 			if phi == 0:
# 				ln_P_phi_given_z[phi,z] = ( -(J_ij_tmp - mu_tmp)**2/(2* (sd_tmp**2) ) ) - 0.5*log_custom(2*np.pi*sd_tmp*sd_tmp) + log_custom(p_tmp)
# 			else:
# 				if J_ij_tmp*(z*2-1) < 0:
# 					ln_P_phi_given_z[phi,z] = 0
# 					bool_tmp = True
# 				else:
# 					ln_P_phi_given_z[phi,z] = ( -( ln_J_ij_tmp - (mu_tmp))**2/(2* (sd_tmp**2) ) ) - 0.5*log_custom(2*np.pi*sd_tmp*sd_tmp) + log_custom(p_tmp)
#
# 		if not bool_tmp:
# 			ln_r = ln_P_phi_given_z[0,z]-ln_P_phi_given_z[1,z]
# 			if ln_r > 5e1:
# 				ln_r = 5e1
# 			else:
# 				if ln_r < -5e1:
# 					ln_r = -5e1
# 			r = exp_custom( ln_r )
# 			P_phi_given_z[0,z] = r/(1+r)
# 			P_phi_given_z[1,z] = 1-P_phi_given_z[0,z]
# 		else:
# 			P_phi_given_z[0,z] = 1-0.01
# 			P_phi_given_z[1,z] = 0+0.01
#
# 	P_phi_tmp = np.zeros(2)
# 	for i in range(2):
# 		P_phi_tmp[i] = P_phi_given_z[i,0]*gamma + P_phi_given_z[i,1]*(1-gamma)
#
# 	P_phi_tmp /= P_phi_tmp.sum()
# 	return P_phi_tmp
#


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def gradients_G_J_ij_jit(input_J_grad, input_H_grad, S, m_i, avg_s_ij_d, which_ij):
	no_R = len(S)
	no_T = len(S[0])-1
	no_N = len(S[0,0,:])
	update_i = which_ij[0]
	update_j = which_ij[1]
	tmp_tanh = np.zeros( (no_R,no_T+1) )
	for r in range(no_R):
		tmp_tanh[r] = np.tanh( np.dot(input_J_grad[update_i,:] , S[r].T)  + input_H_grad[r, update_i])

	gradient_J_ij = 0
	for r in range(no_R):
		gradient_J_ij += no_T*(avg_s_ij_d[r, update_i, update_j] - (S[r,1:,update_j] * tmp_tanh[r,1:]).mean())

	return gradient_J_ij


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def gradients_G_J_ij_jit_prob(input_J_grad, input_H_grad, unique_S_pattern_size, unique_S_pattern, unique_S_pattern_counts, avg_s_ij_d, which_ij):
	no_R = len(unique_S_pattern)
	no_T = int(unique_S_pattern_counts[0].sum()-1)
	no_N = len(unique_S_pattern[0,0,:])
	update_i = which_ij[0]
	update_j = which_ij[1]


	J_vec = input_J_grad[update_i,:]

	grad_tmp1 = 0
	grad_tmp2 = 0
	for r in range(no_R):
		grad_tmp1 += no_T*avg_s_ij_d[r, update_i, update_j]
		patternsize_tmp = unique_S_pattern_size[r]
		H_tmp = input_H_grad[r, update_i]
		pattern_tmp = np.ascontiguousarray(unique_S_pattern[r, :patternsize_tmp, :])
		grad_tmp2 += (unique_S_pattern_counts[r, :patternsize_tmp]*pattern_tmp[:,update_j]*tanh_custom( np.dot(J_vec , pattern_tmp.T)  + H_tmp)).sum()
		# grad_tmp2 += np.dot(unique_S_pattern_counts[r, :patternsize_tmp]*pattern_tmp[:,update_j], np.tanh( np.dot(J_vec , pattern_tmp.T)  + H_tmp))


	gradient_J_ij = grad_tmp1 - grad_tmp2

	return gradient_J_ij


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def log_likelihood_jit_prob(input_J_grad, input_H_grad, unique_S2_pattern_size, unique_S2_pattern, unique_S2_pattern_counts, avg_s_ij_d, which_ij):
	no_R = len(unique_S2_pattern)
	no_T = int(unique_S2_pattern_counts[0].sum()-1)
	no_N = len(unique_S2_pattern[0,0,:])//2
	update_i = which_ij[0]
	update_j = which_ij[1]


	J_vec = input_J_grad[update_i,:]

	grad_tmp1 = 0
	grad_tmp2 = 0
	for r in range(no_R):

		patternsize_tmp = unique_S2_pattern_size[r]
		H_tmp = input_H_grad[r, update_i]
		pattern_tmp = np.ascontiguousarray(unique_S2_pattern[r, :patternsize_tmp, :])
		caltmp = np.dot(J_vec , np.ascontiguousarray(pattern_tmp[:,no_N:].T) )  + H_tmp
		grad_tmp1 += (unique_S2_pattern_counts[r, :patternsize_tmp]*pattern_tmp[:,update_i]*( caltmp )  ).sum()
		grad_tmp2 += (unique_S2_pattern_counts[r, :patternsize_tmp]* log_custom( 2*np.cosh( caltmp ))  ).sum()

	gradient_J_ij = grad_tmp1 - grad_tmp2

	return gradient_J_ij



@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def log_prior_term_jit_prob(input_J_grad, P_z_j, P_phi_ij_given_z_safe, priors_J, small_epsilon, which_ij):

	update_i = which_ij[0]
	update_j = which_ij[1]

	mus = np.array([0, priors_J[1], priors_J[2] ])
	sds = np.array([small_epsilon, priors_J[3]+small_epsilon, priors_J[4]+small_epsilon ])

	mu_phi_z = np.zeros((2,2))
	sd_phi_z = np.zeros((2,2))
	for phi in [0,1]:
		for z in [0,1]:
			mu_phi_z[phi,z] = mus[phi*(1+z)]
			sd_phi_z[phi,z] = sds[phi*(1+z)]

	log_prior = 0
	J_ij_tmp = input_J_grad[update_i, update_j]
	for phi in [0,1]:
		for z in [0,1]:
			if phi == 0:
				caltmp = -(J_ij_tmp**2)/(2*(sd_phi_z[phi,z]**2)+1e-200)
				log_prior += P_z_j[z]*P_phi_ij_given_z_safe[phi,z]*caltmp
			else:
				if J_ij_tmp*(z*2-1)>0:
					caltmp = -( log_custom(np.abs(J_ij_tmp)) - (mu_phi_z[phi,z]) )**2/( 2*(sd_phi_z[phi,z]**2 ) + 1e-200)
				else:
					caltmp = -1e100
				log_prior += P_z_j[z]*P_phi_ij_given_z_safe[phi,z]*caltmp

	return log_prior



@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def gradients_G_J_ij_MF_jit(input_J_grad, input_H_grad, S, m_i, C, D, which_ij):
	no_R = len(S)
	no_T = len(S[0])-1
	no_N = len(S[0,0,:])
	update_i = which_ij[0]
	update_j = which_ij[1]

	D_ij = (D[update_i,update_j])
	C_lj_vec = np.copy(C[:,update_j])

	gradient_J_ij = D_ij - (1-m_i[update_i]**2)* np.dot( input_J_grad[update_i, :], C_lj_vec)

	return gradient_J_ij


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def gradients_G_H_i_jit(input_J_grad, input_H_grad, S, m_i, avg_s_ij_d, which_phase, which_i):
	no_R = len(S)
	no_T = len(S[0])
	no_N = len(S[0,0,:])
	update_r = which_phase
	update_i = which_i

	tmp_tanh = np.tanh( input_J_grad[update_i,:] @ S[update_r].T + input_H_grad[update_r, update_i] ).mean()
	gradients_H_i = no_T*(m_i[update_r, update_i] - tmp_tanh)

	return gradients_H_i


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def gradients_G_H_i_jit_prob(input_J_grad, input_H_grad, m_i, unique_S_pattern_size, unique_S_pattern, unique_S_pattern_counts, which_phase, which_i):
	no_R = len(unique_S_pattern)
	no_T = int(unique_S_pattern_counts[0].sum()-1)
	no_N = len(unique_S_pattern[0,0,:])
	update_r = which_phase
	update_i = which_i

	J_vec = np.ascontiguousarray(input_J_grad[update_i,:])

	grad_tmp1 = no_T * m_i[update_r, update_i]

	patternsize_tmp = unique_S_pattern_size[update_r]
	pattern_tmp = np.ascontiguousarray(unique_S_pattern[update_r, :patternsize_tmp, :])
	pattern_tmp_2 = np.ascontiguousarray(unique_S_pattern_counts[update_r, :patternsize_tmp])
	grad_tmp2 = np.dot(pattern_tmp_2, tanh_custom( np.dot(J_vec , pattern_tmp.T)  + input_H_grad[update_r, update_i]))


	gradients_H_i = grad_tmp1 - grad_tmp2

	return gradients_H_i


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def gradient_ln_prior_J_ij_jit_simple(input_J_grad, input_z_j, input_phi_ij, priors_J, small_epsilon, which_ij):

	which_i = which_ij[0]
	which_j = which_ij[1]

	mus = np.array([priors_J[1], priors_J[2] ])
	sds = np.array([priors_J[3]+small_epsilon, priors_J[4]+small_epsilon ])

	gradient_ln_PJ = 0
	J_ij_tmp =  np.float64(input_J_grad[which_i, which_j])
	z_j_tmp = int(input_z_j[which_j])
	if which_i == which_j:
		z_j_tmp = 0
	phi_ij_tmp = int(input_phi_ij[which_i, which_j])


	# if phi_ij_tmp == 0:
	# 	gradient_ln_PJ =  -J_ij_tmp/(small_epsilon**2+1e-100)
	# else:
	# 	if J_ij_tmp*(z_j_tmp*2-1)>0:
	# 		gradient_ln_PJ =  ( log_custom(np.abs(J_ij_tmp)) - mus[z_j_tmp] )/( (z_j_tmp*2-1) * J_ij_tmp* (sds[z_j_tmp]**2 ) + 1e-100)
	# 	else:
	# 		gradient_ln_PJ = 0

	for z in np.arange(2):
		if J_ij_tmp*(z*2-1)>0:
			gradient_ln_PJ  -=  ( z_j_tmp*z + (1-z_j_tmp)*(1-z) )*( log_custom(np.abs(J_ij_tmp)) - mus[z] )/( (z*2-1) * J_ij_tmp* (sds[z]**2 ) + 1e-100)

	gradient_ln_PJ *= phi_ij_tmp
	gradient_ln_PJ -=  (1-phi_ij_tmp)*J_ij_tmp/(small_epsilon**2+1e-100)

	return gradient_ln_PJ


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def gradient_ln_prior_z_j_jit_simple(input_J_grad, input_z_j, input_phi_ij, priors_J, small_epsilon, which_j):

	gamma_tmp = priors_J[0]
	mus = np.array([priors_J[1], priors_J[2] ])
	sds = np.array([priors_J[3]+small_epsilon, priors_J[4]+small_epsilon ])
	vars = sds**2

	gradient_ln_Pz = log_custom( gamma_tmp/(1-gamma_tmp) )
	J_ij_tmp = input_J_grad[:, which_j]
	z_j_tmp = input_z_j[which_j]
	phi_ij_tmp = input_phi_ij[:, which_j]

	for i in np.arange(len(J_ij_tmp)):
		caltmp = 0
		if (i != which_j):
			for z in np.arange(2):
				if J_ij_tmp[i]*(z*2-1)>0:
					caltmp +=  (-1)*(z*2-1)*( -1*( log_custom(np.abs(J_ij_tmp[i])) - mus[z] )**2/( 2* (vars[z] ) + 1e-100) - 0.5*log_custom(2*np.pi*vars[z])  )
			caltmp*=phi_ij_tmp[i]
		gradient_ln_Pz += caltmp

	return gradient_ln_Pz


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def gradient_ln_prior_phi_j_jit_simple(input_J_grad, input_z_j, input_phi_ij, priors_J, priors_omega, small_epsilon, which_ij):

	which_i = which_ij[0]
	which_j = which_ij[1]


	mus = np.array([priors_J[1], priors_J[2] ])
	sds = np.array([priors_J[3]+small_epsilon, priors_J[4]+small_epsilon ])
	vars = sds**2

	var_epsilon = small_epsilon**2

	gradient_ln_phi = log_custom( (1-priors_omega[which_i, which_j])/priors_omega[which_i, which_j] )

	J_ij_tmp = input_J_grad[which_i, which_j]
	z_j_tmp = input_z_j[which_j]
	phi_ij_tmp = input_phi_ij[which_i, which_j]

	caltmp = 0
	for z in np.arange(2):
		if J_ij_tmp*(z*2-1)>0:
			caltmp += ( z_j_tmp*z + (1-z_j_tmp)*(1-z) )*( -1*( log_custom(np.abs(J_ij_tmp)) - mus[z] )**2/( 2* (vars[z] ) + 1e-100) - 0.5*log_custom(2*np.pi*vars[z])  )

	caltmp -= -1*(J_ij_tmp**2)/(2*var_epsilon) - 0.5*log_custom(2*np.pi*var_epsilon)

	return gradient_ln_phi + caltmp


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def gradient_ln_prior_H_i_jit(input_J_grad, input_H_grad, priors_omega, priors_J, priors_H, small_epsilon, which_phase, which_i):

	update_phase = which_phase
	update_i = which_i

	mean_h = priors_H[which_phase]
	sd_h = priors_H[len(priors_H)//2+which_phase]

	gradient_ln_PH_i = -(input_H_grad[update_phase, update_i] - mean_h)/(sd_h**2+1e-200)

	return gradient_ln_PH_i

@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def ln_prior_H_i_jit(input_J_grad, input_H_grad, priors_omega, priors_J, priors_H, small_epsilon, which_phase, which_i):

	update_phase = which_phase
	update_i = which_i

	mean_h = priors_H[which_phase]
	sd_h = priors_H[len(priors_H)//2+which_phase]

	gradient_ln_PH_i = -(input_H_grad[update_phase, update_i] - mean_h)**2/(2*sd_h**2+1e-200)

	return gradient_ln_PH_i



@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def optimize_J_H( S, R, T, N, m_i, avg_s_ij_d, avg_s_ij_equal_time, unique_S_pattern, unique_S_pattern_size, unique_S_pattern_counts, unique_S2_pattern, unique_S2_pattern_size, unique_S2_pattern_counts, priors_J, priors_H, priors_a, priors_distance, priors_theta, decimated_z_j_states, decimated_phi_ij_states, input_J, input_H, MLE_or_MAP, input_prior_trigger, small_epsilon_for_delta_0, abs_max_J_and_H ):
# def optimize_J_H( S, R, T, N, m_i, avg_s_ij_d, avg_s_ij_equal_time,
# priors_J, priors_H, priors_a, priors_distance, priors_theta,
# decimated_z_j_states, decimated_phi_ij_states,
# input_J, input_H, MLE_or_MAP, input_prior_trigger,
# small_epsilon_for_delta_0, abs_max_J_and_H
# ):

	# priors_J = (Priors[0]) # priors_J[i][j][0:5]: 0->\gamma+ ; 1-> mu_+; 3-> sd_+; 2->mu_-; 4->sd_-
	# priors_H = np.array(Priors[1])
	# priors_a = Priors[2]
	# priors_distance = Priors[3] # the probability of a link is "dead"
	# priors_theta = Priors[4]
	priors_omega = 1-priors_theta*exp_custom(-priors_a*priors_distance)

	e_negative_a_l_ij = exp_custom(-priors_a*priors_distance)

	M_i = np.zeros( N )
	D_ij = np.zeros( (N,N) )
	C_ij = np.zeros( (N,N) )
	for r in range(R):
		M_i += m_i[r]/R
		# for i in range(N):
		# 	for j in range(N):
		# 		D_ij[i][j] += (avg_s_ij_d[r,i,j]-m_i[r,i]*m_i[r,j])/R
		# 		C_ij[i][j] += (avg_s_ij_equal_time[r,i,j]-m_i[r,i]*m_i[r,j])/R
		caltmp = np.outer(m_i[r,:],m_i[r,:])
		D_ij += (avg_s_ij_d[r]-caltmp)/R
		C_ij += (avg_s_ij_equal_time[r]-caltmp)/R


	MLE_J = np.copy(input_J)
	MLE_H = np.copy(input_H)

	prev_J = np.copy(input_J)
	update_J = np.copy(prev_J)
	prev_H = np.copy(input_H)
	update_H = np.copy(prev_H)

	with_self_loop = 1

	# eta_vec = np.array([10.0**i for i in np.arange(-10,1)])
	eta_vec = np.array([10, 1, 0.1, 0.01, 0.001])


	eta_base = 1
	reset_eta = eta_base
	eta = eta_base

	half_life_for_eta = 400
	half_life_for_delta_0 = 50

	# half_life = half_life_for_eta * (N**2 )
	half_life = half_life_for_eta
	half_life_term = (0.5)**(1/half_life)

	time_start_half_life = 0 #1e2 * (N**2 + R*N)

	delta_size = small_epsilon_for_delta_0
	eta_for_delta_function = abs_max_J_and_H
	# half_life_for_delta = half_life_for_delta_0 * (N**2)
	half_life_for_delta = half_life_for_delta_0
	half_life_term_for_delta = (0.5)**(1/half_life_for_delta)

	J_H_move_per_step = 0.05


	timer = 0
	converge_counter = 0
	decimated_J = np.zeros((N,N), dtype=np.bool_)
	converge_counter_J = np.zeros((N,N), dtype=np.int16)
	decimated_H = np.zeros( (R,N) , dtype=np.bool_)
	converge_counter_H = np.zeros( (R,N) , dtype=np.int16)

	decimated_phi = np.zeros((N,N), dtype=np.bool_)
	converge_counter_phi = np.zeros((N,N), dtype=np.int16)
	decimated_z = np.zeros( N , dtype=np.bool_)
	converge_counter_z = np.zeros( N , dtype=np.int16)

	# threshold = 1e-2
	threshold = 1e-5
	converged_lim = 200*(N**2 + R*N + input_prior_trigger*(N**2 + N))
	converged_lim_each = 400

	if_decimation = 1
	prior_trigger = input_prior_trigger

	print(prior_trigger)

	un_deci = N**2
	rand_list = np.arange(un_deci)

	P_old_z_j = np.zeros( (N,2) )
	P_old_phi_ij = np.zeros( (N,N,2,2) )
	z_j = np.zeros(N)
	phi_ij = np.zeros((N,N))
	if(prior_trigger):

		for i in range(N):
			if(decimated_z_j_states[i] != -1):
				P_old_z_j[ i, decimated_z_j_states[i] ] = 1
				decimated_z[i] = 1
			else:
				P_old_z_j[i,:] = np.copy( P_z_j_one_j(input_J, None, priors_theta, e_negative_a_l_ij,  (priors_J), i, small_epsilon_for_delta_0) )

			z_j[i] = np.argmax(P_old_z_j[i,:])

		for i in range(N):
			for j in range(N):
				if (int(decimated_phi_ij_states[ i , j ]) != -1):
					P_old_phi_ij[i,j, int(decimated_phi_ij_states[ i , j ]) , : ] = 1
					decimated_phi[i,j] = 1
				else:
					P_old_phi_ij[i,j,:,:] = np.copy(P_phi_ij_given_z(input_J, priors_theta, e_negative_a_l_ij,  (priors_J), np.array([i,j]) , small_epsilon_for_delta_0))
				phi_ij[i,j] = np.argmax( np.array([(P_old_phi_ij[i,j,p,:] * P_old_z_j[j,:]).sum() for p in [0,1]]) )
	else:
		decimated_z[:] = 1
		decimated_phi[:,:] = 1

	update_phi = np.copy(phi_ij)
	prev_phi = np.copy(update_phi)
	update_z = np.copy(z_j)
	prev_z = np.copy(update_z)

	for i in np.arange(N):
		for j in np.arange(N):
			if decimated_phi[i,j]==1 and decimated_phi_ij_states[ i , j ] ==0:
				decimated_J[i,j] = 1
				update_J[i,j] = 0
			if decimated_z[j] == 1 and (decimated_phi_ij_states[ i , j ]*2-1)*update_J[i,j]<0:
				update_J[i,j] = 0

	# gradients_J_vec = np.copy(update_J)
	# gradients_H_vec = np.copy(update_H)
	# gradients_phi_vec = np.copy(update_phi)
	# gradients_z_vec = np.copy(update_z)
	print(update_z)
	while not( np.all(decimated_J) and np.all(decimated_H) and np.all(decimated_phi) and np.all(decimated_z) ):
		prev_J = np.copy(update_J)
		prev_H = np.copy(update_H)

		prev_phi = np.copy(update_phi)
		prev_z = np.copy(update_z)

		# gradients_J_vec[:,:] = 0
		# gradients_H_vec[:,:] = 0
		# gradients_phi_vec[:,:] = 0
		# gradients_z_vec[:] = 0


		if timer > time_start_half_life:
			eta_base = eta_base*half_life_term
			eta_vec *= half_life_term
			eta = eta_base

		if prior_trigger and eta_for_delta_function > delta_size:
			eta_for_delta_function = eta_for_delta_function*half_life_term_for_delta

		J_to_check = np.argwhere(~decimated_J)

		for update_ij in J_to_check:
			# update_ij = np.array([i, j])
			gradient_J_ij_caltmp = 0

			if prior_trigger:
				# gradients_ln_Prior = 0
				gradients_ln_Prior = gradient_ln_prior_J_ij_jit_simple(prev_J, prev_z, prev_phi, priors_J, eta_for_delta_function, update_ij)
				gradient_J_ij_caltmp += gradients_ln_Prior/T

			if boolean_if_prob:
				gradients_ln_likelihood = gradients_G_J_ij_jit_prob(prev_J, prev_H, unique_S_pattern_size, unique_S_pattern, unique_S_pattern_counts, avg_s_ij_d, update_ij)
			else:
				gradients_ln_likelihood = gradients_G_J_ij_jit(prev_J, prev_H, S, m_i, avg_s_ij_d, update_ij)

			gradient_J_ij_caltmp += gradients_ln_likelihood / (T*R)
			gradient_J_ij_caltmp *= eta
			if gradient_J_ij_caltmp > J_H_move_per_step:
				gradient_J_ij_caltmp = J_H_move_per_step
			else:
				if gradient_J_ij_caltmp < -J_H_move_per_step:
					gradient_J_ij_caltmp = -J_H_move_per_step

			update_tmp = prev_J[update_ij[0], update_ij[1]] + gradient_J_ij_caltmp
			if update_tmp>abs_max_J_and_H:
				update_tmp = abs_max_J_and_H
			else:
				if update_tmp<-abs_max_J_and_H:
					update_tmp = -abs_max_J_and_H

			if (  prior_trigger and (update_ij[0]!=update_ij[1] and decimated_z[update_ij[1]] and (prev_z[update_ij[1]]*2-1)*update_tmp < 0) ):
				update_tmp = 0

			update_J[update_ij[0], update_ij[1]] = update_tmp

			# if timer > 10000:
			# 	print(prev_J[update_ij[0], update_ij[1]], update_J[update_ij[0], update_ij[1]])

			if (not prior_trigger) or eta_for_delta_function<delta_size:
				if update_tmp*prev_J[update_ij[0], update_ij[1]] != 0:
					caltmp_J = np.abs(update_tmp - prev_J[update_ij[0], update_ij[1]] ) / np.abs(prev_J[update_ij[0], update_ij[1]] + 1e-10)
				else:
					caltmp_J = np.abs(update_tmp - prev_J[update_ij[0], update_ij[1]] )

				if caltmp_J < threshold :
					converge_counter += 1
					converge_counter_J[update_ij[0], update_ij[1]] += 1
					if if_decimation and converge_counter_J[update_ij[0], update_ij[1]] > converged_lim_each and decimated_J[update_ij[0], update_ij[1]] == 0:
						decimated_J[update_ij[0], update_ij[1]] = 1
				else:
					converge_counter = 0
					converge_counter_J[update_ij[0], update_ij[1]] = 0


		H_to_check = np.argwhere(~decimated_H)
		for update_ri in H_to_check:
			update_phase = update_ri[0]
			update_i = update_ri[1]
			if boolean_if_prob:
				gradient_h_i_tmp = gradients_G_H_i_jit_prob(prev_J, prev_H, m_i, unique_S_pattern_size, unique_S_pattern, unique_S_pattern_counts, update_phase, update_i) / (T)
			else:
				gradient_h_i_tmp = gradients_G_H_i_jit(prev_J, prev_H, S, m_i, avg_s_ij_d, update_phase, update_i) / (T)

			if(prior_trigger):
				gradient_h_i_tmp += gradient_ln_prior_H_i_jit(prev_J, prev_H, priors_omega, priors_J, priors_H, eta_for_delta_function, update_phase, update_i) / (T)
			gradient_h_i_tmp *= eta
			if gradient_h_i_tmp > J_H_move_per_step:
				gradient_h_i_tmp = J_H_move_per_step
			else:
				if gradient_h_i_tmp < -J_H_move_per_step:
					gradient_h_i_tmp = -J_H_move_per_step
			update_tmp = prev_H[update_phase, update_i] + gradient_h_i_tmp
			if update_tmp>abs_max_J_and_H:
				update_tmp = abs_max_J_and_H
			else:
				if update_tmp<-abs_max_J_and_H:
					update_tmp = -abs_max_J_and_H

			update_H[update_phase, update_i] = update_tmp

			if (not prior_trigger) or eta_for_delta_function<delta_size:
				caltmp_H = np.abs((update_tmp - prev_H[update_phase, update_i] )/(prev_H[update_phase, update_i] + 1e-10))
				if caltmp_H < threshold :
					converge_counter += 1
					converge_counter_H[update_phase, update_i] += 1
					# if if_decimation and converge_counter_H[update_phase, update_i] > converged_lim_each and np.all(decimated_J) and decimated_H[update_phase, update_i] == 0:
					if if_decimation and converge_counter_H[update_phase, update_i] > converged_lim_each and decimated_H[update_phase, update_i] == 0:
						decimated_H[update_phase, update_i] = 1
				else:
					converge_counter = 0
					converge_counter_H[update_phase, update_i] = 0


		if prior_trigger:
			phi_to_check = np.argwhere(~decimated_phi)
			for update_ij in phi_to_check:

				delta_Q_delta_phi = gradient_ln_prior_phi_j_jit_simple(prev_J, prev_z, prev_phi, priors_J, priors_omega, eta_for_delta_function, update_ij)
				delta_Q_delta_phi *= eta
				update_tmp = prev_phi[update_ij[0], update_ij[1]] + delta_Q_delta_phi
				if update_tmp > 1:
					update_tmp = 1
				if update_tmp < 0:
					update_tmp = 0
				update_phi[update_ij[0], update_ij[1]] = update_tmp

				if eta_for_delta_function<delta_size:
					caltmp_phi = update_tmp - prev_phi[update_ij[0], update_ij[1]]
					if caltmp_phi < threshold :
						converge_counter += 1
						converge_counter_phi[update_ij[0], update_ij[1]] += 1
						if if_decimation and converge_counter_phi[update_ij[0], update_ij[1]] > converged_lim_each and (not decimated_phi[update_ij[0], update_ij[1]]) :
							decimated_phi[update_ij[0], update_ij[1]] = 1
							update_phi[update_ij[0], update_ij[1]] = np.round(update_phi[update_ij[0], update_ij[1]])
							if update_phi[update_ij[0], update_ij[1]] == 0:
								decimated_J[update_ij[0], update_ij[1]] = 1
								update_J[update_ij[0], update_ij[1]] = 0
					else:
						converge_counter = 0
						converge_counter_phi[update_ij[0], update_ij[1]] = 0


			z_to_check = np.argwhere(~decimated_z)
			for update_j_array in z_to_check:
				update_j = update_j_array[0]

				delta_Q_delta_z = gradient_ln_prior_z_j_jit_simple(prev_J, prev_z, prev_phi, priors_J, eta_for_delta_function, update_j)
				delta_Q_delta_z *= eta
				update_tmp = prev_z[update_j] + delta_Q_delta_z
				if update_tmp > 1:
					update_tmp = 1
				if update_tmp < 0:
					update_tmp = 0
				update_z[update_j] = update_tmp


				if eta_for_delta_function<delta_size:
					caltmp_z = update_tmp - prev_z[update_j]
					if caltmp_z < threshold :
						converge_counter += 1
						converge_counter_z[update_j] += 1
						if if_decimation and converge_counter_z[update_j] > converged_lim_each and decimated_z[update_j] == 0:
							decimated_z[update_j] = 1
							update_z[update_j] = np.round(update_z[update_j])
							for i in range(len(update_z)):
								if ( decimated_J[i, update_j] and i!=update_j and (update_z[update_j]*2-1)*update_J[i , update_j ]<0  ):
									update_J[i , update_j ] = 0
					else:
						converge_counter = 0
						converge_counter_z[update_j] = 0


		timer += 1

		if timer%(100)==0:
			print(timer, eta, decimated_J.sum(), decimated_H.sum(), decimated_phi.sum(), decimated_z.sum())
			# print(converge_counter_J.flatten())
			# print(decimated_H.flatten())
			# print(decimated_phi.flatten())
			# print(decimated_z.flatten())


	print(update_z)
	print('optimize_J_H:', timer/(N**2+R*N), eta)
	MLE_J = np.copy(update_J)
	MLE_H = np.copy(update_H)
	return update_J, update_H, MLE_J, MLE_H, update_phi, update_z



def MAP_smooth_prior(data, Priors, MLE_or_MAP, if_ML_II, startingJH = None):

	# small_epsilon = 0.02
	small_epsilon_for_delta_0 = 0.05
	abs_max_J_and_H = 10

	def optimize_hyperparameters(input_J_grad, input_H_grad, input_phi_grad, input_z_grad):

		priors_distance2 = priors_distance**2

		# decimation_priors = [gamma, mu1, mu2, sd1, sd2, muH, sdH, a, theta]
		# prev_paras = np.array( [prev_a, prev_gamma, prev_mu_1, prev_sd_1, prev_mu_2, prev_sd_2] )
								# [0:gamma, 1:mu_minus, 2:sd_minus, 3:mu_plus, 4:sd_plus, 5:a, 6:theta ]

		new_priors_H = np.copy(priors_H)
		no_of_phases = len(new_priors_H)//2
		for r in np.arange(no_of_phases):
			if not decimation_priors[len(priors_J) + r]:
				new_priors_H[r] = input_H_grad[r].mean()
			else:
				new_priors_H[r] = np.copy(priors_H[r])
			if not decimation_priors[len(priors_J) + no_of_phases + r]:
				new_priors_H[no_of_phases + r] = ((input_H_grad[r] - new_priors_H[r])**2).mean()
				# new_priors_H[no_of_phases + r] = max(new_priors_H[no_of_phases + r], min_prior_J_sd)
			else:
				new_priors_H[no_of_phases + r] = np.copy(priors_H[no_of_phases + r])

		new_priors_J = np.copy(priors_J)
		ln_J_values = log_custom(np.abs(input_J_grad)+1e-200)
		for z in np.arange(2):
			input_J_grad_caltmp = np.copy(input_J_grad)
			input_J_grad_caltmp[np.diag_indices(len(input_J_grad))] = 0
			choose_J = (input_J_grad_caltmp*(z*2-1) > 0)
			sum_index = np.ones((N,N), dtype=np.bool_)*((input_phi_grad.astype(int))*(input_z_grad == z))*choose_J
			# print(z)
			# print(sum_index)
			mu_tmp = ln_J_values[sum_index].sum()/ (sum_index.sum()+ 1e-10)
			sd_tmp = (((ln_J_values[sum_index]-mu_tmp)**2).sum()/ (sum_index.sum() + 1e-10) )**0.5
			if not decimation_priors[1+z]:
				new_priors_J[1+z] = mu_tmp
			else:
				new_priors_J[1+z] = np.copy(priors_J[1+z])
			if not decimation_priors[3+z]:
				# sd_tmp = max(sd_tmp, min_prior_J_sd)
				new_priors_J[3+z] = sd_tmp
			else:
				new_priors_J[3+z] = np.copy(priors_J[3+z])

		if not decimation_priors[0]:
			new_priors_J[0] = (input_z_grad == 0).sum()/len(input_z_grad)
		else:
			new_priors_J[0] = np.copy(priors_J[0])

		def dQ_da(input_a):
			eal = exp_custom(input_a*priors_distance)
			return -((input_phi_grad*priors_distance).sum()) + ((1-input_phi_grad)*(priors_distance*priors_theta / (eal - priors_theta))).sum()

		if not decimation_priors[ len(priors_J)+len(priors_H) ]:
			startint_a = 0.01
			while dQ_da(startint_a) < 0:
				# print(startint_a, dQ_da(startint_a))
				startint_a /= 2

			startint_b = 2
			while dQ_da(startint_b) > 0:
				# print(startint_b, dQ_da(startint_b))
				startint_b*=2

			new_priors_a = scipy.optimize.bisect(dQ_da, startint_a, startint_b)

			# new_priors_a = min(1, new_priors_a)
		else:
			new_priors_a = np.copy(priors_a)

		e_neg_a_l_ij = exp_custom(-1*new_priors_a*priors_distance)

		P_old_z_j = np.zeros( (N,2) )
		for i in range(N):
			P_old_z_j[i,:] = np.copy( P_z_j_one_j(input_J_grad, None, priors_theta, e_neg_a_l_ij, new_priors_J, i, small_epsilon_for_delta_0) )

		P_old_phi_ij = np.zeros( (N,N,2,2) )
		for i in range(N):
			for j in range(N):
				P_old_phi_ij[i,j,:,:] = np.copy(P_phi_ij_given_z(input_J_grad, priors_theta, e_neg_a_l_ij, new_priors_J, np.array([i,j]) , small_epsilon_for_delta_0))


		# print(P_old_z_j)
		return new_priors_J, new_priors_H, new_priors_a, P_old_z_j, P_old_phi_ij

	S = np.copy(data)

	R = len(S)
	T = len(S[0])
	N = len(S[0,0,:])

	boolean_pass_analysis = False
	if checkpointing:
		checkpoint_file = 'checkpoint_S_statistics.joblib'
		if os.path.exists(checkpoint_file):
			print('Checkpoint exists! Loading S checkpint')
			loaded = load(checkpoint_file)
			unique_S_pattern = np.copy(loaded['unique_S_pattern'])
			unique_S_pattern_size = np.copy(loaded['unique_S_pattern_size'])
			unique_S_pattern_counts = np.copy(loaded['unique_S_pattern_counts'])
			unique_S2_pattern = np.copy(loaded['unique_S2_pattern'])
			unique_S2_pattern_size = np.copy(loaded['unique_S2_pattern_size'])
			unique_S2_pattern_counts = np.copy(loaded['unique_S2_pattern_counts'])
			avg_s_ij_d = np.copy(loaded['avg_s_ij_d'])
			avg_s_ij_equal_time = np.copy(loaded['avg_s_ij_equal_time'])
			m_i = np.copy(loaded['m_i'])
			boolean_pass_analysis = True

	if not boolean_pass_analysis:
		unique_S_pattern = np.zeros(( R,len(S[0]), N ))
		unique_S_pattern_size = np.zeros( R , dtype=int)
		unique_S_pattern_counts = np.zeros( ( R,len(S[0]) ) )

		for r in range(R):
			tmpUniqueS, tmpcounts = np.unique(S[r], axis=0, return_counts=True)
			unique_S_pattern_size[r] = len(tmpUniqueS)
			unique_S_pattern[r, :unique_S_pattern_size[r], :] = np.copy(tmpUniqueS)
			unique_S_pattern_counts[r, :unique_S_pattern_size[r]] = np.copy(tmpcounts)

		unique_S_pattern = unique_S_pattern[:, :int(np.max(unique_S_pattern_size)), :]
		unique_S_pattern_counts = unique_S_pattern_counts[:, :int(np.max(unique_S_pattern_size))]



		S2 = np.zeros((R,len(S[0])-1, 2*N))
		for r in range(R):
			S2[r, :, :N] = S[r,1:,:]
			S2[r, :, N:] = S[r,:len(S[0])-1,:]

		unique_S2_pattern = np.zeros(( R,len(S[0])-1, 2*N ))
		unique_S2_pattern_size = np.zeros( R , dtype=int)
		unique_S2_pattern_counts = np.zeros( ( R,len(S[0])-1 ) )
		for r in range(R):
			tmpUniqueS, tmpcounts = np.unique(S2[r], axis=0, return_counts=True)
			unique_S2_pattern_size[r] = len(tmpUniqueS)
			unique_S2_pattern[r, :unique_S2_pattern_size[r], :] = np.copy(tmpUniqueS)
			unique_S2_pattern_counts[r, :unique_S2_pattern_size[r]] = np.copy(tmpcounts)

		unique_S2_pattern = unique_S2_pattern[:, :int(np.max(unique_S2_pattern_size)), :]
		unique_S2_pattern_counts = unique_S2_pattern_counts[:, :int(np.max(unique_S2_pattern_size))]

		avg_s_ij_d = np.zeros((R,N,N))
		# for i in range(N):
		# 	for j in range(N):
		# 		for r in range(R):
		# 			for t in range(T-1):
		# 				avg_s_ij_d[r][i][j] += S[r][t+1][i]*S[r][t][j]
		# 			avg_s_ij_d[r][i][j] /= (T-1)
		for r in range(R):
			for l in range(unique_S2_pattern_size[r]):
				avg_s_ij_d[r] += unique_S2_pattern_counts[r, l] * np.outer(unique_S2_pattern[r, l, :N], unique_S2_pattern[r, l, N:])
		avg_s_ij_d /= (T - 1)

		avg_s_ij_equal_time = np.zeros((R,N,N))
		# for i in range(N):
		# 	for j in range(N):
		# 		for r in range(R):
		# 			for t in range(T-1):
		# 				avg_s_ij_equal_time[r][i][j] += S[r][t][i]*S[r][t][j]
		# 			avg_s_ij_equal_time[r][i][j] /= (T-1)
		for r in range(R):
			for l in range(unique_S_pattern_size[r]):
				avg_s_ij_equal_time[r] += unique_S_pattern_counts[r, l] * np.outer(unique_S_pattern[r,l], unique_S_pattern[r,l])
		avg_s_ij_equal_time /= T

		m_i = np.zeros((R,N))
		for r in range(R):
			for i in range(N):
				m_i[r,i] = S[r,:,i].mean()

		if checkpointing:
			checkpoint = {
				'unique_S_pattern': unique_S_pattern,
				'unique_S_pattern_size': unique_S_pattern_size,
				'unique_S_pattern_counts': unique_S_pattern_counts,
				'unique_S2_pattern': unique_S2_pattern,
				'unique_S2_pattern_size': unique_S2_pattern_size,
				'unique_S2_pattern_counts': unique_S2_pattern_counts,
				'avg_s_ij_d': avg_s_ij_d,
				'avg_s_ij_equal_time': avg_s_ij_equal_time,
				'm_i': m_i
			}
			dump(checkpoint, checkpoint_file)
			print('saved S checkpoint')

	print(T, unique_S_pattern.shape, unique_S_pattern_size)
	print(T-1, unique_S2_pattern.shape, unique_S2_pattern_size)

	priors_J = Priors[0] # priors_J[i][j][0:5]: 0->\gamma+ ; 1-> mu_+; 3-> sd_+; 2->mu_-; 4->sd_-
	priors_H = Priors[1]
	priors_a = Priors[2]
	priors_distance = Priors[3] # the probability of a link is "dead"
	priors_theta = Priors[4]
	priors_omega = 1-priors_theta*exp_custom(-priors_a*priors_distance)

	min_prior_J_sd = 0.1
	min_prior_H_sd = 0.1
	# small_epsilon_for_delta_0 = min_prior_J_sd

	np.random.seed(10)
	skipMLE = 0
	if startingJH == None:
		start_J = 0.1*(np.random.randn(N,N))
		for i in range(len(start_J)):
			start_J[i,i] = -abs(start_J[i,i])
		start_H = 0.1*(np.random.randn(R,N))
	else:
		start_J = np.copy(startingJH[0])
		start_H = np.copy(startingJH[1])
		if MLE_or_MAP == 'MAP':
			skipMLE = 1


	threshold_for_P_z = 0.9
	decimated_z_j = np.zeros( N ).astype(bool)
	decimated_z_j_states = (np.ones( N )*-1).astype(int)
	prev_z_j_states = np.copy(decimated_z_j_states)
	up_z_j_states = np.copy(decimated_z_j_states)
	P_z_j_converge_counter = np.zeros( N ).astype(int)
	P_z_j_converge_bound = 10

	threshold_for_P_phi = 0.95
	decimated_phi_ij = np.zeros( (N,N) ).astype(bool)
	decimated_phi_ij_states = (np.ones( (N,N) )*-1).astype(int)
	prev_phi_ij_states = np.copy(decimated_phi_ij_states)
	up_phi_ij_states = np.copy(decimated_phi_ij_states)
	P_phi_ij_converge_counter = np.zeros( (N,N) ).astype(int)
	P_phi_ij_converge_bound = 10


	if MLE_or_MAP != 'MAP':
		# print('start_MLE')
		prev_J, prev_H, MLE_J, MLE_H, prev_phi_ij_states, prev_z_j_states = optimize_J_H( S, R, T, N, m_i, avg_s_ij_d, avg_s_ij_equal_time,
																			unique_S_pattern, unique_S_pattern_size, unique_S_pattern_counts,
																			unique_S2_pattern, unique_S2_pattern_size, unique_S2_pattern_counts,
																			priors_J, priors_H, priors_a, priors_distance, priors_theta,
																			decimated_z_j_states, decimated_phi_ij_states,
																			start_J, start_H, 'MLE', 0,
																			small_epsilon_for_delta_0, abs_max_J_and_H )
	else:
		MLE_J = np.copy( start_J )
		prev_J = np.copy( start_J )
		MLE_H = np.copy( start_H )
		prev_H = np.copy( start_H )

		P_old_z_j = np.zeros( (N,2) )
		P_old_phi_ij = np.zeros( (N,N,2,2) )
		z_j = np.zeros(N, dtype=np.int16)
		phi_ij = np.zeros((N,N), dtype=np.int16)

		e_negative_a_l_ij = exp_custom(-priors_a*priors_distance)

		for i in range(N):
			if(decimated_z_j_states[i] != -1):
				P_old_z_j[ i, decimated_z_j_states[i] ] = 1
				decimated_z[i] = 1
			else:
				P_old_z_j[i,:] = np.copy( P_z_j_one_j(MLE_J, None, priors_theta, e_negative_a_l_ij,  (priors_J), i, small_epsilon_for_delta_0) )

			z_j[i] = np.argmax(P_old_z_j[i,:])

		for i in range(N):
			for j in range(N):
				if (int(decimated_phi_ij_states[ i , j ]) != -1):
					P_old_phi_ij[i,j, int(decimated_phi_ij_states[ i , j ]) , : ] = 1
					decimated_phi[i,j] = 1
				else:
					P_old_phi_ij[i,j,:,:] = np.copy(P_phi_ij_given_z(MLE_J, priors_theta, e_negative_a_l_ij,  (priors_J), np.array([i,j]) , small_epsilon_for_delta_0))
				phi_ij[i,j] = np.argmax( np.array([(P_old_phi_ij[i,j,p,:] * P_old_z_j[j,:]).sum() for p in [0,1]]) )

		up_z_j_states = np.copy(z_j)
		up_phi_ij_states = np.copy(phi_ij)

		print(([*priors_J, *priors_H, priors_a]))


	def delta_bound_vec(delt):
		step = step_size_for_priors
		output = np.copy(delt)
		for i in range(len(delt)):
			if delt[i] > step:
				output[i] =  step
			if delt[i] < -step:
				output[i] = -step
		return output

	def delta_bound(delt):
		step = step_size_for_priors
		if delt > step:
			return step
		if delt < -step:
			return -step
		return delt

	update_J = np.copy(prev_J)
	update_H = np.copy(prev_H)

	up_priors_J = np.copy(priors_J)
	up_priors_H = np.copy(priors_H)
	up_priors_a = np.copy(priors_a)
	up_priors_theta = np.copy(priors_theta)


	eta = 1
	eta_base = 1
	half_life = 2
	half_life_term = (0.5)**(1/half_life)

	step_size_for_priors = 0.1

	timer = 0
	converge_counter = 0
	threshold = 1e-3
	converged_lim = 10

	decimation_priors = np.zeros(len(priors_J)+len(priors_H)+1)
	decimation_priors[0] = 1

	# if MLE_or_MAP == 'MAP' and if_ML_II!=1:
	# 	decimation_priors[:] = 1
	decimation_priors_counter = np.zeros(len(decimation_priors))
	new_priors = np.array([*priors_J, *priors_H, priors_a])
	if MLE_or_MAP == 'MAP':
		update_J = np.copy(MLE_J)
		update_H = np.copy(MLE_H)

		if checkpointing and if_ML_II==1:
			checkpoint_file = 'checkpoint_ML2_steps.joblib'
			if os.path.exists(checkpoint_file):
				loaded = load(checkpoint_file)
				timer = np.copy(loaded['timer'])
				decimated_z_j = np.copy(loaded['decimated_z_j'])
				decimated_phi_ij = np.copy(loaded['decimated_phi_ij'])
				decimation_priors = np.copy(loaded['decimation_priors'])
				update_J = np.copy(loaded['update_J'])
				update_H = np.copy(loaded['update_H'])
				priors_J = np.copy(loaded['priors_J'])
				priors_H = np.copy(loaded['priors_H'])
				priors_a = np.copy(loaded['priors_a'])
				up_z_j_states = np.copy(loaded['up_z_j_states'])
				up_phi_ij_states = np.copy(loaded['up_phi_ij_states'])
				eta_base = np.copy(loaded['eta_base'])
				eta = np.copy(loaded['eta'])
				small_epsilon_for_delta_0 = np.copy(loaded['small_epsilon_for_delta_0'])
				priors_omega = np.copy(loaded['priors_omega'])
				P_phi_ij_converge_counter = np.copy(loaded['P_phi_ij_converge_counter'])
				decimated_phi_ij_states = np.copy(loaded['decimated_phi_ij_states'])
				P_z_j_converge_counter = np.copy(loaded['P_z_j_converge_counter'])
				decimated_z_j_states = np.copy(loaded['decimated_z_j_states'])
				decimation_priors_counter = np.copy(loaded['decimation_priors_counter'])
				print('Checkpoint exists! Loading ML2 checkpoint: step ', timer)

		# while timer == 0 or ( converge_counter < converged_lim and not( np.all(decimation_priors) ) ):
		while timer == 0 or ( not((np.all(decimated_z_j) and np.all(decimated_phi_ij)) and np.all(decimation_priors)) ):

			# prev_J = np.copy(update_J)
			# prev_H = np.copy(update_H)

			prev_J = np.copy(start_J)
			prev_H = np.copy(start_H)


			prev_priors_J = np.copy(priors_J)
			prev_priors_H = np.copy(priors_H)
			prev_priors_a = np.copy(priors_a)

			prev_z_j_states = np.copy(up_z_j_states)
			prev_phi_ij_states = np.copy(up_phi_ij_states)


			eta_base *= half_life_term
			eta = eta_base

			round_dp = 3

			# small_epsilon_for_delta_0 = min_prior_J_sd

			up_priors_J, up_priors_H, up_priors_a, P_z_j, P_phi_ij = optimize_hyperparameters(prev_J, prev_H, prev_phi_ij_states, prev_z_j_states)
			# print(P_z_j)

			up_priors_J = np.round(np.copy(up_priors_J),round_dp)
			up_priors_H = np.round(np.copy(up_priors_H),round_dp)
			up_priors_a = np.round(np.copy(up_priors_a),round_dp)

			if timer == 0:
				priors_J = up_priors_J
				priors_H = up_priors_H
				priors_a = up_priors_a
			else:
				priors_J = np.round(prev_priors_J + eta*delta_bound_vec((up_priors_J - prev_priors_J)) ,round_dp)
				priors_H = np.round(prev_priors_H + eta*delta_bound_vec((up_priors_H - prev_priors_H)) ,round_dp)
				priors_a = np.round(prev_priors_a + eta*delta_bound((up_priors_a - prev_priors_a)) ,round_dp)
			priors_J[1] = max(0, priors_J[1])
			priors_J[2] = max(0, priors_J[2])
			priors_J[3] = max((min_prior_J_sd), priors_J[3])
			priors_J[3] = min((1), priors_J[3])
			priors_J[4] = max((min_prior_J_sd), priors_J[4])
			priors_J[4] = min((1), priors_J[4])
			priors_a = min(priors_a,1)

			if MLE_or_MAP == 'MAP' and if_ML_II!=1:
				decimation_priors[:] = 1

			priors_omega = 1-priors_theta*exp_custom(-priors_a*priors_distance)

			prev_priors = np.array([*prev_priors_J, *prev_priors_H, prev_priors_a])
			new_priors = np.array([*priors_J, *priors_H, priors_a])
			frac_diff_tmp = (new_priors-prev_priors)/(prev_priors+1e-10)
			id_tmp = frac_diff_tmp>threshold
			decimation_priors_counter = decimation_priors_counter + 1
			decimation_priors_counter[id_tmp] = 0
			decimation_priors[decimation_priors_counter>converged_lim] = 1

			# for i in range(N):
			# 	up_z_j_states_tmp[i] = int(np.argmax(P_z_j[i,:]))
			# 	tmp_prob = P_z_j[ i, up_z_j_states_tmp[i] ]
			# 	if(tmp_prob>threshold_for_P_z and prev_z_j_states[i] == up_z_j_states_tmp[i]):
			# 		P_z_j_converge_counter[i] += 1
			# 		if P_z_j_converge_counter[i]>=P_z_j_converge_bound:
			# 			decimated_z_j[i] = 1
			# 			decimated_z_j_states[i] = up_z_j_states_tmp[i]
			# 	else:
			# 		P_z_j_converge_counter[i] = 0


			# for i in range(N):
			# 	for j in range(N):
			# 		P_phi_ij_tmp = np.array([(P_phi_ij[i,j,p,:] * P_z_j[j,:]).sum() for p in [0,1]])
			# 		up_phi_ij_states[i,j] = int(np.argmax( P_phi_ij_tmp ))
			# 		tmp_prob = P_phi_ij_tmp[up_phi_ij_states[i,j]]
			# 		if(tmp_prob>threshold_for_P_phi and prev_phi_ij_states[i,j] == up_phi_ij_states[i,j]) and decimated_z_j[j] == 1:
			# 			P_phi_ij_converge_counter[i,j] += 1
			# 			if P_phi_ij_converge_counter[i,j]>=P_phi_ij_converge_bound :
			# 				decimated_phi_ij[i,j] = 1
			# 				decimated_phi_ij_states[i,j] = up_phi_ij_states[i,j]
			# 		else:
			# 			P_phi_ij_converge_counter[i,j] = 0


			if np.all(decimation_priors) :
				for i in range(N):
					for j in range(N):
						if (decimated_phi_ij[i,j] == 0):
							decimated_phi_ij[i,j] = 1
							decimated_phi_ij_states[i,j] = up_phi_ij_states[i,j]
				for i in range(N):
					if (decimated_z_j[i] == 0):
						decimated_z_j[i] = 1
						decimated_z_j_states[i] = up_z_j_states[i]
				# if np.all(decimated_z_j):
				# 	for i in range(N):
				# 		for j in range(N):
				# 			if (decimated_phi_ij[i,j] == 0):
				# 				decimated_phi_ij[i,j] = 1
				# 				decimated_phi_ij_states[i,j] = up_phi_ij_states[i,j]
				# else:
				# 	for i in range(N):
				# 		if (decimated_z_j[i] == 0):
				# 			decimated_z_j[i] = 1
				# 			decimated_z_j_states[i] = up_z_j_states[i]


			up_J, up_H, MLE_J, MLE_H, up_phi_ij_states, up_z_j_states = optimize_J_H( S, R, T, N, m_i, avg_s_ij_d, avg_s_ij_equal_time,
																			unique_S_pattern, unique_S_pattern_size, unique_S_pattern_counts,
																			unique_S2_pattern, unique_S2_pattern_size, unique_S2_pattern_counts,
																			priors_J, priors_H, priors_a, priors_distance, priors_theta,
																			decimated_z_j_states, decimated_phi_ij_states,
																			prev_J, prev_H, 'MAP', 1,
																			small_epsilon_for_delta_0, abs_max_J_and_H )

			for i in range(N):
				if prev_z_j_states[i] == up_z_j_states[i]:
					P_z_j_converge_counter[i] += 1
					if P_z_j_converge_counter[i]>=P_z_j_converge_bound and decimated_z_j[i]==0:
						decimated_z_j[i] = 1
						decimated_z_j_states[i] = up_z_j_states[i]
				else:
					P_z_j_converge_counter[i] = 0
				# print(up_z_j_states[i], P_z_j_converge_counter[i])


			for i in range(N):
				for j in range(N):
					if( prev_phi_ij_states[i,j] == up_phi_ij_states[i,j]) and decimated_z_j[j] == 1 :
						P_phi_ij_converge_counter[i,j] += 1
						if P_phi_ij_converge_counter[i,j]>=P_phi_ij_converge_bound and decimated_phi_ij[i,j] == 0:
							decimated_phi_ij[i,j] = 1
							decimated_phi_ij_states[i,j] = up_phi_ij_states[i,j]
					else:
						P_phi_ij_converge_counter[i,j] = 0

			update_J = up_J
			update_H = up_H




			caltmp_J = np.abs((update_J - prev_J))
			caltmp_J[prev_J!=0] = caltmp_J[prev_J!=0]/prev_J[prev_J!=0]
			converged_J = np.zeros( ( len(prev_J), len(prev_J) ) )
			converged_J[ caltmp_J < threshold] = 1

			caltmp_H = np.abs((update_H - prev_H))
			caltmp_H[prev_H!=0] = caltmp_H[prev_H!=0]/prev_H[prev_H!=0]
			converged_H = np.zeros( prev_H.shape )
			converged_H[ caltmp_H < threshold] = 1

			if( (np.all(converged_H) and np.all(converged_J)) or np.all(decimation_priors) ):
				converge_counter += 1
			else:
				converge_counter = 0

			timer += 1

			if checkpointing and if_ML_II==1:
				checkpoint_file = 'checkpoint_ML2_steps.joblib'
				checkpoint = {
					'timer': timer,
					'decimated_z_j': decimated_z_j,
					'decimated_phi_ij': decimated_phi_ij,
					'decimation_priors': decimation_priors,
					'update_J': update_J,
					'update_H': update_H,
					'priors_J': priors_J,
					'priors_H': priors_H,
					'priors_a': priors_a,
					'up_z_j_states': up_z_j_states,
					'up_phi_ij_states': up_phi_ij_states,
					'eta_base': eta_base,
					'eta': eta,
					'small_epsilon_for_delta_0': small_epsilon_for_delta_0,
					'priors_omega': priors_omega,
					'P_phi_ij_converge_counter': P_phi_ij_converge_counter,
					'decimated_phi_ij_states': decimated_phi_ij_states,
					'P_z_j_converge_counter': P_z_j_converge_counter,
					'decimated_z_j_states': decimated_z_j_states,
					'decimation_priors_counter': decimation_priors_counter,

				}
				dump(checkpoint, checkpoint_file)
				print('saved ML_2 checkpoint: step ', timer)

			print(timer, decimation_priors.sum(), len(decimation_priors)-decimation_priors.sum(), eta, new_priors, decimated_z_j.sum(), decimated_phi_ij.sum(), sep='\t')

		print(new_priors)

	return update_J, update_H


def mean_fields(summaries, kinetic_ising_method):

	def matrix_A_by_nMF():
		A_nMF = scaling_beta*np.diag(1-m_i**2 )
		return A_nMF

	def matrix_A_by_TAP():
		caltmp = (scaling_beta**2)*(1-m_i**2) * np.matmul( np.power(J_nMF.copy(),2) , 1 - np.power(m_i,2)   )
		F_i = np.array( [min(np.roots([1,-2,1,-1*caltmp[i]]).real, key=abs) for i in range(len(caltmp))] )
		A_TAP = A_nMF*(1-F_i)
		return A_TAP

	def matrix_h_by_nMF():
		# L = (1/(2*scaling_beta))*log_custom((1+m_i)/(1-m_i))
		# h_nMF = L - np.matmul( J_nMF, m_i )

		h_nMF = np.zeros((R_size, N))
		for r in range(R_size):
			L = (1/(2*scaling_beta))*log_custom((1+m_i_r[r])/(1-m_i_r[r]))
			h_nMF[r] = L - np.dot( J_nMF, m_i_r[r] )

		return h_nMF

	def matrix_h_by_TAP():
		# L = (1/(2*scaling_beta))*log_custom((1+m_i)/(1-m_i))
		# caltmp = m_i * np.matmul( np.power(J_TAP.copy(),2) , 1 - np.power(m_i,2)   )
		# h_TAP = L - np.matmul( J_TAP, m_i ) + scaling_beta*caltmp

		h_TAP = np.zeros((R_size, N))
		for r in range(R_size):
			L = (1/(2*scaling_beta))*log_custom((1+m_i_r[r])/(1-m_i_r[r]))
			caltmp = m_i_r[r] * np.matmul( np.power(J_TAP.copy(),2) , 1 - np.power(m_i_r[r],2)   )
			h_TAP[r] = L - np.matmul( J_TAP, m_i_r[r] ) + scaling_beta*caltmp
		return h_TAP

	input_counter = 0
	m_i_r = summaries[input_counter]; input_counter+=1
	C_r = summaries[input_counter]; input_counter+=1
	D_r = summaries[input_counter]; input_counter+=1
	S_r = summaries[input_counter]; input_counter+=1

	N = len(m_i_r[0])
	R_size = len(m_i_r)
	T_len = np.array([len(S_r[i]) for i in range(len(S_r))])

	m_i = np.zeros(N)
	C = np.zeros((N,N))
	D = np.zeros((N,N))
	for i in range(N):
		for r in range(R_size):
			if m_i_r[r,i] == 1:
				m_i_r[r,i] -= 1e-3
			if m_i_r[r,i] == -1:
				m_i_r[r,i] += 1e-3
			m_i[i] += m_i_r[r,i]*T_len[r]/(T_len.sum())
		for j in range(N):
			for r in range(R_size):
				C[i,j] += C_r[r,i,j]*T_len[r]/(T_len.sum())
				D[i,j] += D_r[r,i,j]*T_len[r]/(T_len.sum())

	scaling_beta = 1

	A_nMF = matrix_A_by_nMF()
	J_nMF = np.linalg.pinv(A_nMF) @ D @ np.linalg.pinv(C)

	# D = (1-m**2) * J * C

	method_counter = 0
	if kinetic_ising_method==method_counter:
		h_nMF = matrix_h_by_nMF()
		return [J_nMF, h_nMF]
	method_counter +=1

	if kinetic_ising_method==method_counter:
		A_TAP = matrix_A_by_TAP()
		J_TAP = np.linalg.pinv(A_TAP) @ D @ np.linalg.pinv(C)
		h_TAP = matrix_h_by_TAP()
		return [J_TAP, h_TAP]
	method_counter +=1




@njit(nogil=True, cache=True, fastmath=True)
def generate_time_series(J_ij_input, H_i_phases, number_of_phase, time_step):
	N = len(J_ij_input)

	number_of_sample = number_of_phase

	m_i = np.zeros((number_of_sample, N))
	C = np.zeros((number_of_sample, N,N))
	D = np.zeros((number_of_sample, N,N))
	S = np.zeros((number_of_sample, time_step, N))

	prob_x = np.zeros(N)

	for r in range(number_of_sample):
		for i in range(N):
			S[r, 0, i] = np.random.randint(2)*2-1
		H_i_input = np.copy(H_i_phases[r])
		for t in range(time_step-1):
			h_vec = H_i_input + np.dot(J_ij_input , S[r, t+0,:])
			caltmp = (2*h_vec)
			for i,v in enumerate(caltmp):
				if v > 40:
					prob_x[i] = 1
				else:
					if v < -40:
						prob_x[i] = 0
					else:
						prob_x[i] = 1/(exp_custom(-v)+1)
			randtmp = np.random.rand(N)
			for i in range(N):
				S[r, t+1, i] = -1
				if randtmp[i] < prob_x[i]:
					S[r, t+1, i] = 1
			m_i[r] += S[r, t+1, :]
			tmpC = np.array( [[(S[r, t+1, i] * S[r, t+1, j]) for j in range(N)] for i in range(N)] )
			C[r] += tmpC
			tmpD = np.array( [[(S[r, t+1, i] * S[r, t+0, j]) for j in range(N)] for i in range(N)] )
			D[r] += tmpD

		for i in range(N):
			m_i[r,i] = m_i[r,i]/(time_step-1)
		for i in range(N):
			for j in range(N):
				C[r,i,j] = C[r,i,j]/((time_step-1)) - m_i[r,i]*m_i[r,j]
				D[r,i,j] = D[r,i,j]/((time_step-1)) - m_i[r,i]*m_i[r,j]

	return m_i, C, D, S


def generate_kin_data( inputs ):

	# inputs = [
	# 10 , # N
	# 2000,
	# 0.2,
	# [[0.2, 0.8], [0.5, 0], [0.1, 0.1]],
	# [[-1, 0, 1], [0.1, 0.1, 0.1]],
	# 0
	# ]

	which_model = 'mix_neurons'

	input_counter = 0
	N = int(inputs[input_counter]); input_counter += 1
	time_step = int(inputs[input_counter]); input_counter += 1

	Network_para = (inputs[input_counter]); input_counter += 1


	Js_ratio = inputs[input_counter][0]
	J_means = inputs[input_counter][1]
	J_sds = inputs[input_counter][2]; input_counter += 1

	number_of_phase = len(inputs[input_counter][0])

	H_means = inputs[input_counter][0]
	H_sds = inputs[input_counter][1]; input_counter += 1

	which_sample_id = int(inputs[input_counter]); input_counter += 1

	np.random.seed(1+which_sample_id)

	Adj_mat, neighbor_mat, connectivity_mat, distance, loc = generate_RRG_with_loc_neural(N, Network_para, 1)


	J_ij = generate_data_guassian_mixture(Js_ratio, J_means, J_sds, N**2)
	J_ij = J_ij.T
	J_ij = Adj_mat*J_ij

	H_i_phases = np.array([np.random.normal(H_means[i], H_sds[i], size=(N)) for i in range(number_of_phase)])

	m_i, C, D, S = generate_time_series(J_ij, H_i_phases, number_of_phase, time_step)

	priors_J = np.array([Js_ratio[0], *J_means, *J_sds])
	priors_H = np.array([*H_means, *H_sds])
	priors_a = Network_para
	priors_distance = np.copy(distance)
	priors_theta = np.ones(J_ij.shape)
	Priors = [priors_J, priors_H, priors_a, priors_distance, priors_theta]

	return J_ij, H_i_phases, m_i, C, D, S, loc, Priors

import networkx as nx
import matplotlib.pyplot as plt
import numpy as np
import matplotlib.cm as cm
import matplotlib
def data_analysis( data_inputs, structure, priors_input = None , solution_J = None, file_name = None):
	def generate_network(J_plot, pos_plot, nx_file_name):

		plt.clf()
		# Create an empty directed graph
		G = nx.DiGraph()

		# Add nodes to the graph with coordinates
		A = np.copy(J_plot)

		num_nodes = len(A)
		coordinates = [(pos_plot[i][0], pos_plot[i][1]) for i in range(num_nodes)]  # Example coordinates, replace with your own
		G.add_nodes_from(range(num_nodes))
		pos_tmp = dict(zip(range(num_nodes), coordinates))

		# Iterate over the adjacency matrix and add edges with appropriate colors
		tmp_edge_colors = []
		for i in range(num_nodes):
			for j in range(num_nodes):
				weight = A[i][j]
				if weight != 0:
					tmp_edge_colors.append((i, j, weight))
					G.add_edge(j, i, weight=weight)

		# Extract the weights and colors as separate lists
		weights = [edge[2] for edge in tmp_edge_colors]

		node_size = 1500 / num_nodes  # Adjust the coefficient as per your preference

		# Calculate the edge widths based on the weight magnitudes
		max_width = 3  # Maximum width for the edges
		min_weight = -np.max(np.abs(weights))
		max_weight = np.max(np.abs(weights))
		normalized_weights = [(weight - min_weight) / (max_weight - min_weight) for weight in weights]

		# Create a colormap
		cmap = matplotlib.colormaps['coolwarm']

		for i, (source, target, weight) in enumerate(tmp_edge_colors):
			# print(weight, min_weight, max_weight, normalized_weights[i])
			edge_color_tmp = np.array(cmap(normalized_weights[i]))
			edge_color_tmp = [c for c in edge_color_tmp]
			width = 3 * normalized_weights[i]  # Adjust the width scaling as desired
			alpha = (1-exp_custom(-10*np.abs(weight))) / (1-exp_custom(-10))
			if alpha > 1:
				alpha = 1
			if alpha < 0:
				alpha = 0
			if source == target:  # Check for self-loop
				tmp_node_size = node_size*0.6
			else:
				tmp_node_size = node_size
			edge_color_tmp[-1] = alpha
			# print(edge_color_tmp)
			nx.draw_networkx_edges(G, pos_tmp, edgelist=[(source, target)], width=width, node_size = tmp_node_size, edge_color=edge_color_tmp )

		sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=min_weight, vmax=max_weight))
		sm.set_array([])
		cbar = plt.colorbar(sm, ax=plt.gca())
		# Plot the graph
		nx.draw_networkx_nodes(G, pos_tmp, node_size=node_size)
		# nx.draw_networkx_labels(G, pos_tmp)
		# nx.draw_networkx_edge_labels(G, pos_tmp, edge_labels=edge_weights)
		plt.axis('off')
		plt.savefig(nx_file_name)
		G.clear()
		plt.close()
		plt.clf()


	def plot_all_graphs(inferred_data, true_data, quantity_name, method_name):
		if not true_data is None:
			plt.close()
			min_lim = min(  inferred_data.min(), true_data.min() )
			max_lim = max(  inferred_data.max(), true_data.max() )
			plt.scatter(true_data.flatten(), inferred_data.flatten(), label=method_name)
			plt.plot([min_lim, max_lim], [min_lim, max_lim], color='k')
			plt.xlim((min_lim,max_lim))
			plt.ylim((min_lim,max_lim))
			plt.ylabel('est {}'.format(quantity_name))
			plt.xlabel('true {}'.format(quantity_name))
			plt.legend()
			plt.savefig('{}_{}_{}_compare.pdf'.format(file_name, quantity_name, method_name))
			plt.close()

			plt.hist(true_data.flatten(), bins=50)
			plt.savefig('{}_{}_true_hist.pdf'.format(file_name, quantity_name))
			plt.close()

			plt.hist(inferred_data.flatten(), bins=50)
			plt.savefig('{}_{}_{}_hist.pdf'.format(file_name, quantity_name, method_name))
			plt.close()


			if len(true_data.shape) >= 3:
				for p in range(true_data.shape[0]):
					plt.close()
					min_lim = min(  inferred_data[p].min(), true_data[p].min() )
					max_lim = max(  inferred_data[p].max(), true_data[p].max() )
					plt.scatter(true_data[p].flatten(), inferred_data[p].flatten(), label=method_name)
					plt.plot([min_lim, max_lim], [min_lim, max_lim], color='k')
					plt.xlim((min_lim,max_lim))
					plt.ylim((min_lim,max_lim))
					plt.ylabel('est {}'.format(quantity_name))
					plt.xlabel('true {}'.format(quantity_name))
					plt.legend()
					plt.savefig('{}_{}_{}_phase_{}_compare.pdf'.format(file_name, quantity_name, method_name, p))
					plt.close()

			if len(true_data.shape) < 3:
				plt.close()
				abs_max = np.abs(true_data).max()
				plt.matshow( true_data , cmap='RdBu_r', vmin=-abs_max, vmax=abs_max )
				plt.savefig('{}_{}_true_mat.pdf'.format(file_name, quantity_name) )
				plt.close()

				np.savetxt(
				'{}_{}_true.txt'.format(file_name, quantity_name),
				true_data,
				fmt='%f',
				delimiter='\t'
				)
			else:
				for p in range(true_data.shape[0]):
					plt.close()
					abs_max = np.abs(true_data[p]).max()
					plt.matshow( true_data[p] , cmap='RdBu_r', vmin=-abs_max, vmax=abs_max )
					plt.savefig('{}_{}_phase_{}_true_mat.pdf'.format(file_name, quantity_name, p) )
					plt.close()

					np.savetxt(
					'{}_{}_phase_{}_true.txt'.format(file_name, quantity_name, p),
					true_data[p],
					fmt='%f',
					delimiter='\t'
					)

		if len(inferred_data.shape) < 3:
			plt.close()
			abs_max = np.abs(inferred_data).max()
			plt.matshow( inferred_data , cmap='RdBu_r', vmin=-abs_max, vmax=abs_max )
			plt.savefig('{}_{}_{}_mat.pdf'.format(file_name, quantity_name, method_name) )
			plt.close()

			plt.hist( inferred_data.flatten() )
			plt.savefig('{}_{}_{}_hist.pdf'.format(file_name, quantity_name, method_name) )
			plt.close()

			np.savetxt(
			'{}_{}_{}.txt'.format(file_name, quantity_name, method_name),
			inferred_data,
			fmt='%f',
			delimiter='\t'
			)
		else:
			for p in range(inferred_data.shape[0]):
				plt.close()
				abs_max = np.abs(inferred_data[p]).max()
				plt.matshow( inferred_data[p] , cmap='RdBu_r', vmin=-abs_max, vmax=abs_max )
				plt.savefig('{}_{}_phase_{}_{}_mat.pdf'.format(file_name, quantity_name, p, method_name) )
				plt.close()

				np.savetxt(
				'{}_{}_phase_{}_{}.txt'.format(file_name, quantity_name, p, method_name),
				inferred_data[p],
				fmt='%f',
				delimiter='\t'
				)

	def get_priors(loc, add_structure):

		def grouping_loc_gen_theta(structure):
			if structure == 0:
				return 0.7*np.ones((len(distance),len(distance)))

			if structure == 1:
				groups_bound_i_x_y = np.zeros((2,2,2,2))
				for i in range(2):
					for j in range(2):
						groups_bound_i_x_y[i,j,0,:] = [0+0.5*(i%2), 0.5+ 0.5*(i%2)]
						groups_bound_i_x_y[i,j,1,:] = [0+0.5*(j%2), 0.5+ 0.5*(j%2)]

				x = loc[:,0]
				y = loc[:,1]

				which_group_ij = np.ones((len(x),2)).astype(int)

				which_group_ij[x<0.5,0] = 0
				which_group_ij[y<0.5,1] = 0

				tmp_theta = np.ones((len(x),len(x)))*0.9
				for i in range(len(x)):
					for j in range(len(x)):
						groups_length = (np.abs(which_group_ij[i,:] - which_group_ij[j,:])).sum()
						tmp_theta[i,j] = tmp_theta[i,j]*(0.4**groups_length)

				return tmp_theta

			if structure == 2:
				x = loc[:,0]
				y = loc[:,1]

				modules = np.zeros( len(x) ,dtype=np.int16)
				tmp_theta = np.ones((len(x),len(x)))*0.9

				stripes_bounds = np.array(Stripes_Bound_input)
				for i in np.arange(1,len(stripes_bounds)-1):
					bound_1 = stripes_bounds[i]
					bound_2 = stripes_bounds[i+1]
					modules[np.where((x > bound_1) & (x<=bound_2))[0]] = int(i)

				for i in range(len(x)):
					for j in range(len(x)):
						groups_length = np.abs( modules[i] - modules[j] )
						tmp_theta[i,j] = tmp_theta[i,j]*(0.5**groups_length)

				return tmp_theta



		N = len(loc)
		posx = np.copy(loc[:,0])
		posy = np.copy(loc[:,1])
		distance = np.array([[ ((posx[i]-posx[j])**2 + (posy[i]-posy[j])**2)**(0.5)  for j in range(N) ] for i in range(N)])
		ratio_0 = 0.2
		J_mean_0 = 0
		J_mean_1 = 0
		J_sd = 1
		H_mean = -0.5
		H_sd = 1
		prior_a = 0.001
		prior_theta = grouping_loc_gen_theta(add_structure)

		prior_J_tmp = np.array([ratio_0, J_mean_0, J_mean_1, J_sd, J_sd])
		prior_H_tmp = np.array([*(np.ones(R)*H_mean), *(np.ones(R)*H_sd)])

		return [np.array(prior_J_tmp), np.array(prior_H_tmp), prior_a, distance, prior_theta]

	# for testing
	# data_inputs = [m_i, C, D, S, loc]
	# priors_input = Priors

	m_i, C, D, S, loc = data_inputs
	R = len(S)
	true_data = [solution_J , None, m_i, C, D]
	if priors_input == None:
		priors_J, priors_H, priors_a, priors_distance, priors_theta = get_priors(loc, structure)
	else:
		priors_J, priors_H, priors_a, priors_distance, priors_theta = priors_input
	Priors = priors_J, priors_H, priors_a, priors_distance, priors_theta

	quantity_name_list = ['J_est', 'H_est', 'm_i_est', 'C_est', 'D_est']
	quantity_plot = np.array([0,0,1,1,1], dtype=bool)
	if not solution_J is None:
		quantity_plot[0] = 1
	quantity_plot = np.arange(len(quantity_name_list))[quantity_plot]

	boolean_pass_analysis = False
	if checkpointing:
		checkpoint_file = 'checkpoint_nMF.joblib'
		if os.path.exists(checkpoint_file):
			print('Checkpoint exists! Loading nMF checkpint')
			loaded = load(checkpoint_file)
			J_est = np.copy(loaded['J_est'])
			H_est = np.copy(loaded['H_est'])
			boolean_pass_analysis = True

	if not boolean_pass_analysis:
		inputs_MF = [m_i, C, D, S]
		print('start nMF')
		J_est, H_est = mean_fields(inputs_MF, 0)
		if checkpointing:
			checkpoint = {
				'J_est': J_est,
				'H_est': H_est
			}
			dump(checkpoint, checkpoint_file)
			print('saved nMF checkpoint')
		print('end inference nMF')
		m_i_est, C_est, D_est, S_est = generate_time_series(J_est, H_est, len(H_est), 10000)
		print('end fitting nMF')
		inferred_nMF = [J_est, H_est, m_i_est, C_est, D_est]
		# generate_network(J_est, loc, '{}_{}_{}.pdf'.format(file_name, 'network', 'nMF'))
		for j,q in enumerate(quantity_name_list):
			plot_all_graphs(inferred_nMF[j], true_data[j], q, 'nMF')


	# ratio_tmp = 0.2 / np.abs(J_est).max()
	# startingJH = [J_est*ratio_tmp, 0.1*(np.random.randn( *H_est.shape ))]

	startingJH = [J_est, H_est ]

	boolean_pass_analysis = False
	if checkpointing:
		checkpoint_file = 'checkpoint_MLE.joblib'
		if os.path.exists(checkpoint_file):
			print('Checkpoint exists! Loading MLE checkpint')
			loaded = load(checkpoint_file)
			J_est = np.copy(loaded['J_est'])
			H_est = np.copy(loaded['H_est'])
			boolean_pass_analysis = True

	if not boolean_pass_analysis:
		print('start MLE')
		J_est, H_est = MAP_smooth_prior(S, Priors, 'MLE', 0, startingJH )
		if checkpointing:
			checkpoint = {
				'J_est': J_est,
				'H_est': H_est
			}
			dump(checkpoint, checkpoint_file)
			print('saved MLE checkpoint')

		print('end inference MLE')
		m_i_est, C_est, D_est, S_est = generate_time_series(J_est, H_est, len(H_est), 10000)
		print('end fitting MLE')
		inferred_MLE = [J_est, H_est, m_i_est, C_est, D_est]
		# generate_network(J_est, loc, '{}_{}_{}.pdf'.format(file_name, 'network', 'MLE'))
		for j,q in enumerate(quantity_name_list):
			plot_all_graphs(inferred_MLE[j], true_data[j], q, 'MLE')

	startingJH = [J_est, H_est]

	boolean_pass_analysis = False
	if checkpointing:
		checkpoint_file = 'checkpoint_MAP.joblib'
		if os.path.exists(checkpoint_file):
			print('Checkpoint exists! Loading MAP checkpint')
			loaded = load(checkpoint_file)
			J_est = np.copy(loaded['J_est'])
			H_est = np.copy(loaded['H_est'])
			boolean_pass_analysis = True

	if not boolean_pass_analysis:
		print('start MAP')
		J_est, H_est = MAP_smooth_prior(S, Priors, 'MAP', 0, startingJH )
		if checkpointing:
			checkpoint = {
				'J_est': J_est,
				'H_est': H_est
			}
			dump(checkpoint, checkpoint_file)
			print('saved MAP checkpoint')
		print('end inference MAP')
		m_i_est, C_est, D_est, S_est = generate_time_series(J_est, H_est, len(H_est), 10000)
		print('end fitting MAP')
		inferred_MAP = [J_est, H_est, m_i_est, C_est, D_est]
		# generate_network(J_est, loc, '{}_{}_{}.pdf'.format(file_name, 'network', 'MAP'))
		for j,q in enumerate(quantity_name_list):
			plot_all_graphs(inferred_MAP[j], true_data[j], q, 'MAP')

	# startingJH = [J_est, H_est]

	# print('start ML-II')
	# J_est, H_est = MAP_smooth_prior(S, Priors, 'MAP', 1, startingJH )
	# print('end inference ML-II')
	# m_i_est, C_est, D_est, S_est = generate_time_series(J_est, H_est, len(H_est), 10000)
	# print('end fitting ML-II')
	# inferred_MLII = [J_est, H_est, m_i_est, C_est, D_est]
	# # generate_network(J_est, loc, '{}_{}_{}.pdf'.format(file_name, 'network', 'ML-II'))
	# for j,q in enumerate(quantity_name_list):
	# 	plot_all_graphs(inferred_MLII[j], true_data[j], q, 'ML-II')


	# method_name_list = ['nMF', 'MLE', 'MAP', 'ML-II']
	# inferred_data = [inferred_nMF, inferred_MLE, inferred_MAP, inferred_MLII]
	#
	# for i,m in enumerate(method_name_list):
	# 	for j,q in enumerate(quantity_name_list):
	# 		plot_all_graphs(inferred_data[i][j], true_data[j], q, m)

	return 0
	# return inferred_nMF, inferred_MLE, inferred_MAP, inferred_MLII

	# method_name_list = ['nMF']
	# inferred_data = [inferred_nMF]
	#
	# for i,m in enumerate(method_name_list):
	# 	for j,q in enumerate(quantity_name_list):
	# 		plot_all_graphs(inferred_data[i][j], true_data[j], q, m)
	#
	# return inferred_nMF

def modules_grouping(loc, structure):
	if rescale_pos_ROI:
		loc = np.copy((loc-loc.min(axis=0))/(loc.max(axis=0)-loc.min(axis=0)))
	posx = np.copy(loc[:,0])
	posy = np.copy(loc[:,1])

	N = len(posx)

	distance = np.array([[ ((posx[i]-posx[j])**2 + (posy[i]-posy[j])**2)**(0.5)  for j in range(N) ] for i in range(N)])

	if structure == 0:
		return np.zeros( len(posx) )

	if structure == 1:
		groups_bound_i_x_y = np.zeros((2,2,2,2))
		for i in range(2):
			for j in range(2):
				groups_bound_i_x_y[i,j,0,:] = [0+0.5*(i%2), 0.5+ 0.5*(i%2)]
				groups_bound_i_x_y[i,j,1,:] = [0+0.5*(j%2), 0.5+ 0.5*(j%2)]

		x = loc[:,0]
		y = loc[:,1]

		which_group_ij = np.ones((len(x),2)).astype(int)

		which_group_ij[x<0.5,0] = 0
		which_group_ij[y<0.5,1] = 0

		modules = np.ones( len(posx) )
		for i in range(len(modules)):
			modules[i] = which_group_ij[i,0] + 2*which_group_ij[i,1]

		return modules

	if structure == 2:
		modules = np.zeros( len(posx) ,dtype=np.int16)

		stripes_bounds = np.array(Stripes_Bound_input)
		for i in np.arange(1,len(stripes_bounds)-1):
			bound_1 = stripes_bounds[i]
			bound_2 = stripes_bounds[i+1]
			modules[np.where((posx > bound_1) & (posx<=bound_2))[0]] = int(i)

		return modules

def reorder_neurons_from_dataset(ROI_name, ROI_header, spikes_name, spikes_header, indices_to_keep, tail_name = '.txt', modules=None):
	import pandas as pd
	# fN = 20
	# r = '0.00'
	# perm = 0
	# seed = 0
	# indices_to_keep = np.arange(0,22,2)
	# ROI_name = 'stim_N_{}_ratio_{}_perm_{}_seed_{}_ROIs.txt'.format(fN, r, perm, seed)
	# spikes_name = 'stim_N_{}_ratio_{}_perm_{}_seed_{}_spikes.txt'.format(fN, r, perm, seed)
	# ROI_header = 'infer'
	# spikes_header = None

	ROIs = pd.read_csv(ROI_name, header=ROI_header)
	modules = np.arange(len(ROIs))
	ROIs.columns = ['ROI id', ' x', ' y']
	ROIs['modules'] = modules
	ROIs = ROIs[ROIs['ROI id'].isin(indices_to_keep)]
	ROIs = ROIs.sort_values('modules').reset_index(drop=True).reset_index()
	ROIs = ROIs.rename(columns={'ROI id': 'old ROI id', 'index': 'valid ROI id'})

	spikes_data = pd.read_csv(spikes_name, header=spikes_header)
	spikes_data.columns = ['old_id', 'spike_time']
	spikes_data['new_id'] = np.ones(len(spikes_data)) * np.nan
	for i in np.arange(len(ROIs['old ROI id'])):
		old_id = ROIs['old ROI id'][i]
		new_id = ROIs['valid ROI id'][i]
		spikes_data.loc[ spikes_data['old_id'] == old_id , 'new_id'] = new_id
	spikes_data = spikes_data.dropna()
	spikes_data['new_id'] = spikes_data['new_id'].astype(int)

	ROIs = ROIs[ ['valid ROI id', ' x', ' y', 'modules'] ]
	spikes_data = spikes_data[ ['new_id', 'spike_time'] ]

	new_ROI_name = '{}_newid.txt'.format(ROI_name.split(tail_name)[0])
	ROIs.to_csv(new_ROI_name, index=False)

	new_spikes_name = '{}_newid.txt'.format(spikes_name.split(tail_name)[0])
	spikes_data.to_csv(new_spikes_name, index=False)


	return np.array(spikes_data)

def reorder_neurons_from_dataset_MEA(ROIs_array, spikes_name, spikes_header, indices_to_keep, modules=None):
	import pandas as pd

	ROIs = pd.DataFrame()
	ROIs['ROI id'] = ROIs_array[:,0]
	ROIs[' x'] = ROIs_array[:,1]
	ROIs[' y'] = ROIs_array[:,2]
	ROIs['modules'] = modules
	ROIs = ROIs[ROIs['ROI id'].isin(indices_to_keep)]
	ROIs = ROIs.sort_values('modules').reset_index(drop=True).reset_index()
	ROIs = ROIs.rename(columns={'ROI id': 'old ROI id', 'index': 'valid ROI id'})

	spikes_data = pd.read_csv(spikes_name, header=spikes_header)
	spikes_data.columns = ['old_id', 'spike_time']
	spikes_data['new_id'] = np.ones(len(spikes_data)) * np.nan
	for i in np.arange(len(ROIs['old ROI id'])):
		old_id = ROIs['old ROI id'][i]
		new_id = ROIs['valid ROI id'][i]
		spikes_data.loc[ spikes_data['old_id'] == old_id , 'new_id'] = new_id
	spikes_data = spikes_data.dropna()
	spikes_data['new_id'] = spikes_data['new_id'].astype(int)

	old_new_id_pair = ROIs[ ['old ROI id', 'valid ROI id', ' x', ' y', 'modules'] ]

	ROIs = ROIs[ ['valid ROI id', ' x', ' y', 'modules'] ]
	spikes_data = spikes_data[ ['new_id', 'spike_time'] ]

	old_new_ROI_name = '{}_ROI_old_newid.csv'.format( spikes_name.split('.')[0] )
	old_new_id_pair.to_csv(old_new_ROI_name, index=False)

	new_ROI_name = '{}_ROI_newid.csv'.format( spikes_name.split('.')[0] )
	ROIs.to_csv(new_ROI_name, index=False)

	new_spikes_name = '{}_newid.csv'.format( spikes_name.split('.')[0] )
	spikes_data.to_csv(new_spikes_name, index=False)


	return np.array(spikes_data)

def turn_activities_into_matrix(activities, bin_size, max_n = None, max_t = None):
	if max_t is None:
		t_size = activities[:,1].max()+bin_size
	else:
		t_size = max_t+bin_size
	bins = np.arange(0, t_size, bin_size)
	time_instance = np.digitize(activities[:,1], bins)-1
	neurons = (activities[:,0] - activities[:,0].min()).astype(int)
	if max_n is None:
		n_size = neurons.max()+1
	else:
		n_size = max_n
	S = -1*np.ones( ( len(bins) , n_size) , dtype=np.float64)
	for i,n in enumerate(neurons):
		S[ time_instance[i], n ] = 1

	total_spikes = np.sum( (S+1)//2 , axis=1)
	sparsity = np.sum(total_spikes==0)/len(total_spikes)

	boolean_array = (total_spikes==0)
	result_array = boolean_array[1:] & boolean_array[:-1]
	result_array = np.insert(result_array, 0, False)
	new_S = S[np.invert(result_array),:]

	new_total_spikes = np.sum( (new_S+1)//2 , axis=1)
	new_sparsity = np.sum(new_total_spikes==0)/len(new_total_spikes)

	# return S , sparsity, new_S, new_sparsity
	return new_S


def gross_mutal_infromation_list(activities_list, start_tau, end_tau, step_size):

	tau_list = np.arange(start_tau, end_tau, step_size )

	gross_mutal_infromation_list = np.zeros((len(tau_list),2))

	for ii in range(len(tau_list)):
		delta_Tau = np.round(tau_list[ii],4)
		print(delta_Tau)

		if len(activities_list) == 1:
			S = turn_activities_into_matrix(activities_list[0], delta_Tau)
		else:
			max_n = 0
			max_t = 0
			for r in range(len(activities_list)):
				max_n = max( max_n, int(1+activities_list[r][:,0].max()-activities_list[r][:,0].min() ) )
				max_t = max( max_t, activities_list[r][:,1].max()+delta_Tau )

			S = turn_activities_into_matrix(activities_list[0], delta_Tau, max_n, max_t)
			for r in np.arange(1,len(activities_list)):
				S_tmp = turn_activities_into_matrix(activities_list[r], delta_Tau, max_n, max_t)
				S = np.vstack([S,S_tmp])

		neurons_list = np.arange(len(S[0]))
		m_i_tmp = S.mean(axis=0)
		S = S[:,np.invert(m_i_tmp==-1)]

		no_of_spins = len(S[0])
		M = len(S)
		n_ij_alpha_beta = np.zeros( (no_of_spins, no_of_spins, 2 ,2) )
		index_i1 = S[0]==1
		index_i0 = np.invert(index_i1)
		for t in range(M-1):
			index_j1 = np.copy(index_i1)
			index_j0 = np.copy(index_i0)
			index_i1 = S[t+1]==1
			index_i0 = np.invert(index_i1)
			n_ij_alpha_beta[np.ix_(index_i1, index_j1 ,[1],[1])] +=1
			n_ij_alpha_beta[np.ix_(index_i1, index_j0 ,[1],[0])] +=1
			n_ij_alpha_beta[np.ix_(index_i0, index_j1 ,[0],[1])] +=1
			n_ij_alpha_beta[np.ix_(index_i0, index_j0 ,[0],[0])] +=1

		I_delta_tau = np.zeros( (no_of_spins, no_of_spins) )
		for i in range(no_of_spins):
			for j in range(no_of_spins):
				tmp_I_Dt = 0
				for t0 in [0,1]:
					for t1 in [0,1]:
						caltmp = n_ij_alpha_beta[i][j][t1][t0]/(M-1)
						pi = (n_ij_alpha_beta[i][j][t1][0] + n_ij_alpha_beta[i][j][t1][1])/(M-1)
						pj = (n_ij_alpha_beta[i][j][0][t0] + n_ij_alpha_beta[i][j][1][t0])/(M-1)
						caltmp1 = 0
						if caltmp != 0 and (pi*pj)!=0:
							caltmp1 = caltmp*log_custom( caltmp/(pi*pj) )
						tmp_I_Dt += caltmp1
				I_delta_tau[i][j] = tmp_I_Dt

		gross_mutal_infromation = (M-1)*(I_delta_tau.sum() - I_delta_tau.diagonal().copy().sum())
		gross_mutal_infromation_list[ii, 0] = delta_Tau
		gross_mutal_infromation_list[ii, 1] = gross_mutal_infromation

	return gross_mutal_infromation_list[gross_mutal_infromation_list[:,1].argmax(), 0],  gross_mutal_infromation_list

def Akke_emulator_analysis(fN, ratios_list, perm, seed, add_structure):
	start_tau, end_tau, step_size = 0.01, 0.5, 0.02

	def output_raster_plot(data, name):
		plt.close()
		plt.scatter(data.T[1], data.T[0], marker='.', s=5)
		plt.savefig(name)
		plt.close()
	# fN = 20
	# ratios_list = ['0.00', '1.00']
	# perm = 0
	# seed = 0

	activities_list = []
	loc_list = []
	W_ij_list = []
	for r in ratios_list:
		tmp_name = 'stim_N_{}_ratio_{}_perm_{}_seed_{}_spikes.csv'.format(fN, r, perm, seed)
		activities_list.append( np.loadtxt(tmp_name, delimiter=',') )
		output_raster_plot(activities_list[-1], 'stim_N_{}_ratio_{}_perm_{}_seed_{}_raster.pdf'.format(fN, r, perm, seed)  )
		tmp_name = 'stim_N_{}_ratio_{}_perm_{}_seed_{}_ROIs.csv'.format(fN, r, perm, seed)
		loc_list.append( np.loadtxt(tmp_name, delimiter=',', skiprows=1 ) )
		tmp_name = 'stim_N_{}_ratio_{}_perm_{}_seed_{}_W.csv'.format(fN, r, perm, seed)
		W_ij_list.append( np.loadtxt(tmp_name, delimiter=',' ) )

	# optimal_tau, gross_mut_info = gross_mutal_infromation_list(activities_list, start_tau, end_tau, step_size)

	optimal_tau = 0.02

	max_n = 0
	max_t = 0
	for r in range(len(activities_list)):
		max_n = max( max_n, int(1+activities_list[r][:,0].max()-activities_list[r][:,0].min() ) )
		max_t = max( max_t, activities_list[r][:,1].max()+optimal_tau )
	S = np.array([turn_activities_into_matrix(activities_list[i], optimal_tau, max_n, max_t) for i in range(len(activities_list))])

	m_i, C, D = get_m_C_D_from_S(S)

	if rescale_pos_ROI:
		loc = (loc_list[0][:,1:]-loc_list[0][:,1:].min(axis=0))/(loc_list[0][:,1:].max(axis=0)-loc_list[0][:,1:].min(axis=0))
	else:
		loc = np.copy(loc_list[0][:,1:])

	data_inputs = [m_i, C, D, S, loc]
	file_name = 'stim_N_{}_perm_{}_seed_{}_spikes.csv'.format(fN, perm, seed)
	results = data_analysis( data_inputs, add_structure, None , W_ij_list[0], file_name)

def in_vitro_analysis( phases_name , add_structure, store_file_name):
	start_tau, end_tau, step_size = 0.05, 0.5, 0.05

	def output_raster_plot(data, name):
		plt.close()
		plt.scatter(data.T[1], data.T[0], marker='.', s=5)
		plt.savefig(name)
		plt.close()

	spike_raster_name = '_raster.pdf'
	spike_tail_name = '_spikesData.csv'
	ROI_tail_name = '_ROI_centers_everything.csv'

	# store_file_name = 'C1'
	# phases_name = ['C1_1_spont', 'C1_2_stim']

	activities_list = []
	loc_list = []
	for r in phases_name:
		tmp_name = '{}{}'.format(r, spike_tail_name)
		activities_list.append( np.loadtxt(tmp_name, delimiter=',', skiprows=1 ) )
		output_raster_plot(activities_list[-1], '{}{}'.format(r, spike_raster_name)  )
		tmp_name = '{}{}'.format(r, ROI_tail_name)
		loc_list.append( np.loadtxt(tmp_name, delimiter=',', skiprows=1 ) )

	modules = modules_grouping(loc_list[0][:,1:], add_structure)
	index_to_keep = np.unique( np.concatenate( [np.unique(activities_list[r][:,0].astype(int)) for r in range(len(activities_list))] ) )

	for r in phases_name:
		spikes_name = '{}{}'.format(r, spike_tail_name)
		ROI_name = '{}{}'.format(r, ROI_tail_name)
		reorder_neurons_from_dataset(ROI_name, 'infer', spikes_name, 'infer', index_to_keep, '.csv', modules)

	spike_raster_name = '_raster_newid.pdf'
	spike_tail_name = '_spikesData_newid.txt'
	ROI_tail_name = '_ROI_centers_everything_newid.txt'
	activities_list = []
	loc_list = []
	for r in phases_name:
		tmp_name = '{}{}'.format(r, spike_tail_name)
		activities_list.append( np.loadtxt(tmp_name, delimiter=',', skiprows=1 ) )
		output_raster_plot(activities_list[-1], '{}{}'.format(r, spike_raster_name)  )
		tmp_name = '{}{}'.format(r, ROI_tail_name)
		loc_list.append( np.loadtxt(tmp_name, delimiter=',', skiprows=1 ) )


	optimal_tau, gross_mut_info = gross_mutal_infromation_list(activities_list, start_tau, end_tau, step_size)

	np.savetxt('{}_mutal_info_opt_{}.txt'.format(store_file_name, optimal_tau), gross_mut_info, fmt='%f', delimiter='\t')


	max_n = 0
	max_t = 0
	for r in range(len(activities_list)):
		max_n = max( max_n, int(1+activities_list[r][:,0].max()-activities_list[r][:,0].min() ) )
		max_t = max( max_t, activities_list[r][:,1].max()+optimal_tau )
	S = np.array([turn_activities_into_matrix(activities_list[i], optimal_tau, max_n, max_t) for i in range(len(activities_list))])

	m_i, C, D = get_m_C_D_from_S(S)

	if rescale_pos_ROI:
		loc = (loc_list[0][:,1:]-loc_list[0][:,1:].min(axis=0))/(loc_list[0][:,1:].max(axis=0)-loc_list[0][:,1:].min(axis=0))
	else:
		loc = np.copy(loc_list[0][:,1:])

	data_inputs = [m_i, C, D, S, loc]
	file_name = store_file_name
	results = data_analysis( data_inputs, add_structure, None , None, file_name)


def MEA_analysis( phases_files, title_name, structure_type, structure_file=None, ROI_file=None, opt_tau=None):
	portion_to_keep = [0,1]
	start_tau, end_tau, step_size = 0.001, 0.2, 0.001
	fequency_thresholding = global_frequency_cut
	keep_top_neurons = 0

	def output_raster_plot(data, name):
		plt.close()
		plt.scatter(data.T[1], data.T[0], marker='.', s=5)
		plt.savefig(name)
		plt.close()

	def create_x_y_pos_MEA():
		pos = np.zeros((64**2, 3))
		iter_tmp = 0
		for i in range(64):
			for j in range(64):
				pos[iter_tmp][0] = iter_tmp
				pos[iter_tmp][1] = i/63
				pos[iter_tmp][2] = j/63
				iter_tmp += 1
		return pos

	def kill_activities_by_time_ratio(activities, portion_to_keep):
		begin_time = portion_to_keep[0]*(activities[:,1].max())
		end_time = portion_to_keep[1]*(activities[:,1].max())

		spike_data_update = activities[ activities[:,1] >= begin_time ,:]
		spike_data_update = spike_data_update[ spike_data_update[:,1] <= end_time ,:]

		spike_data_update[:,1] = spike_data_update[:,1] - spike_data_update[:,1].min()

		return spike_data_update

	def kill_activities_by_modules(activities, modules):
		list_to_kill = np.arange(len(modules))[modules == -1]
		spikes = activities[:,0]
		mask = np.invert( np.in1d(spikes, list_to_kill) )
		rows = np.where(mask)[0]
		spike_data_update = activities[rows,:]
		return spike_data_update

	def kill_activities_by_sparsity(activities, hz_threshold):
		unique_id, counts = np.unique(activities[:,0] , return_counts=True)
		time = activities[:,1].max() - activities[:,1].min()
		frequency = counts/(time)
		hz_threshold_tmp = hz_threshold
		if keep_top_neurons > 0:
			sorted_freq = np.sort(frequency)[::-1]
			hz_threshold_tmp = sorted_freq[keep_top_neurons-1]

		list_to_kill = np.array( unique_id[frequency<hz_threshold_tmp] ).astype(int)
		spikes = activities[:,0]
		mask = np.invert( np.in1d(spikes, list_to_kill) )
		rows = np.where(mask)[0]
		spike_data_update = activities[rows,:]
		return spike_data_update

	def transforing_activities(activities, gap_size, new_gap_size):
		event_time = np.copy(activities[:,1])
		event_id = np.copy(activities[:,0])
		new_time = [event_time[0]]
		for i in range(1, len(event_time)):
			if event_time[i] - event_time[i-1] > gap_size:
				event_time[i:] -= (event_time[i] - event_time[i-1]) - new_gap_size
		new_activities = np.copy(activities)
		new_activities[:,1] = event_time
		return new_activities

	spike_raster_name = '_raster.pdf'

	if ROI_file == None:
		loc = create_x_y_pos_MEA()
	else:
		loc = np.loadtxt(ROI_file, delimiter=',', skiprows=1 )
		if rescale_pos_ROI:
			loc[:,1:] = (loc[:,1:]-loc[:,1:].min(axis=0))/(loc[:,1:].max(axis=0)-loc[:,1:].min(axis=0))

	if not structure_file==None:
		ROIs_regions = np.loadtxt(structure_file, delimiter=',', skiprows=1 )
		modules = np.copy(ROIs_regions[:,1])
	else:
		modules = modules_grouping( loc[:,1:] , structure_type)

	activities_list = []
	for r in phases_files:
		activities_tmp = np.loadtxt(r, delimiter=',', skiprows=1 )

		unique_id, counts = np.unique(activities_tmp[:,0] , return_counts=True)
		time = activities_tmp[:,1].max() - activities_tmp[:,1].min()
		frequency = counts/(time)
		plt.close()
		plt.hist(frequency, bins=np.logspace(np.log10(frequency.min()), np.log10(frequency.max()), 20), log=True)
		plt.xscale('log')
		plt.savefig('{}_frequency.pdf'.format(r.split('.')[0]))
		np.savetxt('{}_frequency.txt'.format(r.split('.')[0]), frequency)

		activities_tmp = kill_activities_by_time_ratio(activities_tmp, portion_to_keep)
		activities_tmp = kill_activities_by_modules(activities_tmp, modules)
		activities_tmp = kill_activities_by_sparsity(activities_tmp, fequency_thresholding)
		output_raster_plot(activities_tmp, '{}{}'.format(r.split('.')[0], spike_raster_name)  )
		activities_list.append(activities_tmp)

	index_to_keep = np.unique( np.concatenate( [np.unique(activities_list[r][:,0].astype(int)) for r in range(len(activities_list))] ) )

	# print(index_to_keep)

	for r in phases_files:
		reorder_neurons_from_dataset_MEA(loc, r, 'infer', index_to_keep, modules)


	if opt_tau == None:
		spike_tail_name = '_newid.csv'
		ROI_tail_name = '_ROI_newid.csv'
		activities_list = []
		loc_list = []
		for r in phases_files:
			tmp_name = '{}{}'.format(r.split('.')[0], spike_tail_name)
			activities_tmp = np.loadtxt(tmp_name, delimiter=',', skiprows=1 )
			activities_tmp = kill_activities_by_time_ratio(activities_tmp, portion_to_keep)
			activities_tmp = kill_activities_by_modules(activities_tmp, modules)
			activities_tmp = kill_activities_by_sparsity(activities_tmp, fequency_thresholding)
			activities_list.append( activities_tmp )
			tmp_name = '{}{}'.format(r.split('.')[0], ROI_tail_name)
			loc_list.append( np.loadtxt(tmp_name, delimiter=',', skiprows=1 ) )
		optimal_tau, gross_mut_info = gross_mutal_infromation_list(activities_list, start_tau, end_tau, step_size)
		np.savetxt('{}_mutal_info_opt_{}.txt'.format(title_name, optimal_tau), gross_mut_info, fmt='%f', delimiter='\t')
	else:
		optimal_tau = opt_tau
	# optimal_tau = 0.05

	spike_tail_name = '_newid.csv'
	ROI_tail_name = '_ROI_newid.csv'
	activities_list = []
	loc_list = []
	for r in phases_files:
		tmp_name = '{}{}'.format(r.split('.')[0], spike_tail_name)
		activities_tmp = np.loadtxt(tmp_name, delimiter=',', skiprows=1 )
		activities_tmp = kill_activities_by_time_ratio(activities_tmp, portion_to_keep)
		activities_tmp = kill_activities_by_modules(activities_tmp, modules)
		activities_tmp = kill_activities_by_sparsity(activities_tmp, fequency_thresholding)
		activities_list.append( activities_tmp )
		tmp_name = '{}{}'.format(r.split('.')[0], ROI_tail_name)
		loc_list.append( np.loadtxt(tmp_name, delimiter=',', skiprows=1 ) )


	max_n = 0
	max_t = 0
	for r in range(len(activities_list)):
		max_n = max( max_n, int(1+activities_list[r][:,0].max()-activities_list[r][:,0].min() ) )
		max_t = max( max_t, activities_list[r][:,1].max()+optimal_tau )
	S =([turn_activities_into_matrix(activities_list[i], optimal_tau, max_n, max_t) for i in range(len(activities_list))])
	# S = replace_consecutive_zero_rows(S)
	S = np.array(S)

	m_i, C, D = get_m_C_D_from_S(S)

	if rescale_pos_ROI:
		loc = (loc_list[0][:,1:]-loc_list[0][:,1:].min(axis=0))/(loc_list[0][:,1:].max(axis=0)-loc_list[0][:,1:].min(axis=0))
	else:
		loc = np.copy(loc_list[0][:,1:])

	data_inputs = [m_i, C, D, S, loc]
	file_name = title_name
	results = data_analysis( data_inputs, structure_type, None , None, file_name)

def kinetic_ising_toy_model(inputs):

	# inputs = [
	# 10 , # N
	# 2000,
	# 0.2,
	# [[0.2, 0.8], [0.5, 0], [0.1, 0.1]],
	# [[-2, 0], [0.1, 0.1]],
	# 0
	# ]

	J_ij, H_i, m_i, C, D, S, loc, Priors = generate_kin_data(inputs)
	data_inputs = [m_i, C, D, S, loc]

	inferred_structures = data_analysis(data_inputs,0, None, J_ij, 'toy_model')



def MEA_batch(  phases_title_list, phases_name_list, structure_type, structure_file=None, ROI_file=None, opt_tau=None):

	number_of_dataset = len(phases_title_list)
	for i,title_name in enumerate(phases_title_list):
		if structure_file == None:
			tmp_struc_file = None
		else:
			tmp_struc_file = structure_file[i]
		if ROI_file == None:
			tmp_ROI_file = None
		else:
			tmp_ROI_file = ROI_file[i]
		MEA_analysis( phases_name_list[i], title_name, structure_type, tmp_struc_file, tmp_ROI_file, opt_tau)






# Anna Ca 1
# Stripes_Bound_input = [0,0.08,0.3,0.45,0.66,0.81,1] # stripes for Anna data (experiment 10.1)
# rescale_pos_ROI = True

# Anna Ca 2
# Stripes_Bound_input = [0,0.16,0.346,0.49,0.66,0.81,1] # stripes for Anna data (experiment 16.3)
# rescale_pos_ROI = True

# Akke emu
Stripes_Bound_input = [0, 0.3, 0.5, 0.8, 1] # stripes for Akke data
rescale_pos_ROI = False

boolean_if_prob = True # use prob integrate for gradients

global_frequency_cut = 0

checkpointing = True
delete_after = False

if sys.argv[1] == 'general':
	global_frequency_cut = float(sys.argv[3])

	phases_title_list = [ sys.argv[4] ]

	phases_name_list = [
		sys.argv[5:-1]
	]

	structure_type = int(sys.argv[2]) # 0:homo ; 1:4 islands; 2: stripes

	structure_file = None

	ROI_file = [
	sys.argv[-1]
	]

	# print(phases_title_list)
	# print(phases_name_list)
	# print(structure_type)
	# print(ROI_file)

	MEA_batch( phases_title_list, phases_name_list, structure_type, structure_file, ROI_file, None)










if sys.argv[1] == 'toy':
	inputs = [
	20, # N
	400,
	0.2,
	[[0.2, 0.8], [0.3, 0], [0.1, 0.1]],
	[[0], [0.1]],
	0
	]
	kinetic_ising_toy_model(inputs)
















#
